from abc import ABC, abstractmethod
from typing import Any, TypeVar

from pydantic import BaseModel

from .run_context import RunContext

MemT = TypeVar("MemT", bound="Memory")


class Memory(BaseModel, ABC):
    @abstractmethod
    def reset(
        self, *args: Any, ctx: RunContext[Any] | None = None, **kwargs: Any
    ) -> None:
        pass

    @abstractmethod
    def erase(self) -> None:
        pass

    @abstractmethod
    def update(
        self, *args: Any, ctx: RunContext[Any] | None = None, **kwargs: Any
    ) -> None:
        pass

    @property
    @abstractmethod
    def is_empty(self) -> bool:
        pass

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class DummyMemory(Memory):
    def reset(
        self, *args: Any, ctx: RunContext[Any] | None = None, **kwargs: Any
    ) -> None:
        pass

    def erase(self) -> None:
        pass

    def update(
        self, *args: Any, ctx: RunContext[Any] | None = None, **kwargs: Any
    ) -> None:
        pass

    @property
    def is_empty(self) -> bool:
        return True



from typing import cast

from grasp_agents.errors import CompletionError

from ..typing.completion import Completion, Usage
from . import LiteLLMChoice, LiteLLMCompletion, LiteLLMUsage
from .message_converters import from_api_assistant_message


def from_api_completion_usage(api_usage: LiteLLMUsage) -> Usage:
    reasoning_tokens = None
    cached_tokens = None

    if api_usage.completion_tokens_details is not None:
        reasoning_tokens = api_usage.completion_tokens_details.reasoning_tokens
    if api_usage.prompt_tokens_details is not None:
        cached_tokens = api_usage.prompt_tokens_details.cached_tokens

    input_tokens = api_usage.prompt_tokens - (cached_tokens or 0)
    output_tokens = api_usage.completion_tokens  # - (reasoning_tokens or 0)

    return Usage(
        input_tokens=input_tokens,
        output_tokens=output_tokens,
        reasoning_tokens=reasoning_tokens,
        cached_tokens=cached_tokens,
    )


def from_api_completion(
    api_completion: LiteLLMCompletion, name: str | None = None
) -> Completion:
    if len(api_completion.choices) > 1:
        raise CompletionError("Multiple choices are not supported")
    api_choice = cast("LiteLLMChoice", api_completion.choices[0])

    message = from_api_assistant_message(api_choice.message, name=name)

    api_usage = getattr(api_completion, "usage", None)
    usage: Usage | None = None
    if api_usage:
        usage = from_api_completion_usage(cast("LiteLLMUsage", api_usage))
        hidden_params = getattr(api_completion, "_hidden_params", {})
        usage.cost = hidden_params.get("response_cost")

    return Completion(
        id=api_completion.id,
        model=api_completion.model,
        name=name,
        created=api_completion.created,
        system_fingerprint=api_completion.system_fingerprint,
        message=message,
        finish_reason=api_choice.finish_reason,  # type: ignore[assignment, arg-type]
        logprobs=getattr(api_choice, "logprobs", None),
        usage=usage,
        provider_specific_fields=getattr(api_choice, "provider_specific_fields", None),
        hidden_params=api_completion._hidden_params,  # type: ignore[union-attr]
        response_ms=getattr(api_completion, "_response_ms", None),
    )


def to_api_completion(completion: Completion) -> LiteLLMCompletion:
    raise NotImplementedError



import logging
from collections.abc import AsyncIterator, Sequence
from typing import Any, ClassVar, Generic, cast

from grasp_agents.tracing_decorators import agent

from ..memory import MemT
from ..packet import Packet
from ..run_context import CtxT, RunContext
from ..typing.events import Event, ProcPacketOutputEvent, ProcPayloadOutputEvent
from ..typing.io import InT, OutT, ProcName
from .base_processor import BaseProcessor, with_retry, with_retry_stream

logger = logging.getLogger(__name__)


class MappingProcessor(
    BaseProcessor[InT, OutT, MemT, CtxT], Generic[InT, OutT, MemT, CtxT]
):
    """
    Processor that can have different numbers of inputs and outputs, allowing for an
    arbitrary mapping between them.
    """

    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    async def _process(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: list[InT] | None = None,
        memory: MemT,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> list[OutT]:
        """
        Process a list of inputs and return a list of outputs. The length of the
        output list can be different from the input list.
        """
        return cast("list[OutT]", in_args)

    async def _process_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: list[InT] | None = None,
        memory: MemT,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> AsyncIterator[Event[Any]]:
        outputs = await self._process(
            chat_inputs=chat_inputs,
            in_args=in_args,
            memory=memory,
            call_id=call_id,
            ctx=ctx,
        )
        for output in outputs:
            yield ProcPayloadOutputEvent(
                data=output, proc_name=self.name, call_id=call_id
            )

    def _preprocess(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
        forgetful: bool = False,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> tuple[list[InT] | None, MemT]:
        val_in_args = self._validate_inputs(
            call_id=call_id,
            chat_inputs=chat_inputs,
            in_packet=in_packet,
            in_args=in_args,
        )
        memory = self.memory.model_copy(deep=True) if forgetful else self.memory

        return val_in_args, memory

    def _join_routings(
        self, routings: list[Sequence[ProcName] | None]
    ) -> Sequence[Sequence[ProcName]] | None:
        if all(r is None for r in routings):
            joined_routing = None
        else:
            joined_routing = [r or [] for r in routings]
        return joined_routing

    def _postprocess(
        self, outputs: list[OutT], call_id: str, ctx: RunContext[CtxT]
    ) -> Packet[OutT]:
        payloads: list[OutT] = []
        routings: list[Sequence[ProcName] | None] = []
        for output in outputs:
            val_output = self._validate_output(output, call_id=call_id)
            payloads.append(val_output)

            selected_recipients = self.select_recipients(
                output=val_output, ctx=ctx, call_id=call_id
            )
            routings.append(selected_recipients)

        routing = self._join_routings(routings)

        return Packet(sender=self.name, payloads=payloads, routing=routing)

    @agent(name="processor")  # type: ignore
    @with_retry
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
        forgetful: bool = False,
        call_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> Packet[OutT]:
        ctx = ctx or RunContext[CtxT](state=None)  # type: ignore
        call_id = self._generate_call_id(call_id)

        val_in_args, memory = self._preprocess(
            chat_inputs=chat_inputs,
            in_packet=in_packet,
            in_args=in_args,
            forgetful=forgetful,
            call_id=call_id,
            ctx=ctx,
        )
        outputs = await self._process(
            chat_inputs=chat_inputs,
            in_args=val_in_args,
            memory=memory,
            call_id=call_id,
            ctx=ctx,
        )

        return self._postprocess(outputs=outputs, call_id=call_id, ctx=ctx)

    @agent(name="processor")  # type: ignore
    @with_retry_stream
    async def run_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
        forgetful: bool = False,
        call_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> AsyncIterator[Event[Any]]:
        ctx = ctx or RunContext[CtxT](state=None)  # type: ignore
        call_id = self._generate_call_id(call_id)

        val_in_args, memory = self._preprocess(
            chat_inputs=chat_inputs,
            in_packet=in_packet,
            in_args=in_args,
            forgetful=forgetful,
            call_id=call_id,
            ctx=ctx,
        )
        outputs: list[OutT] = []
        async for event in self._process_stream(
            chat_inputs=chat_inputs,
            in_args=val_in_args,
            memory=memory,
            call_id=call_id,
            ctx=ctx,
        ):
            if isinstance(event, ProcPayloadOutputEvent):
                outputs.append(event.data)
            yield event

        out_packet = self._postprocess(outputs=outputs, call_id=call_id, ctx=ctx)

        yield ProcPacketOutputEvent(
            data=out_packet, proc_name=self.name, call_id=call_id
        )



from collections.abc import Sequence
from typing import Any

from pydantic import Field

from .memory import Memory
from .run_context import RunContext
from .typing.io import LLMPrompt
from .typing.message import Message, Messages, SystemMessage


class LLMAgentMemory(Memory):
    messages: Messages = Field(default_factory=Messages)
    instructions: LLMPrompt | None = Field(default=None)

    def __init__(self, instructions: LLMPrompt | None = None) -> None:
        super().__init__()
        self.reset(instructions)

    def reset(
        self, instructions: LLMPrompt | None = None, ctx: RunContext[Any] | None = None
    ):
        if instructions is not None:
            self.instructions = instructions

        self.messages = (
            [SystemMessage(content=self.instructions)]
            if self.instructions is not None
            else []
        )

    def erase(self) -> None:
        self.messages = []

    def update(
        self, new_messages: Sequence[Message], *, ctx: RunContext[Any] | None = None
    ):
        self.messages.extend(new_messages)

    @property
    def is_empty(self) -> bool:
        return len(self.messages) == 0

    def __repr__(self) -> str:
        return f"LLMAgentMemory with message history of length {len(self.messages)}"



import time
from enum import StrEnum
from typing import Any, Generic, Literal, TypeVar, get_args
from uuid import uuid4

from pydantic import BaseModel, ConfigDict, Field

from ..packet import Packet
from .completion import Completion
from .completion_chunk import (
    AnnotationsChunk,
    CompletionChunk,
    RefusalChunk,
    ResponseChunk,
    ThinkingChunk,
    ToolCallChunk,
)
from .message import (
    AssistantMessage,
    MessageBase,
    SystemMessage,
    ToolCall,
    ToolMessage,
    UserMessage,
)


class EventSourceType(StrEnum):
    LLM = "llm"
    AGENT = "agent"
    USER = "user"
    TOOL = "tool"
    PROC = "processor"
    WORKFLOW = "workflow"
    RUN = "run"


class EventType(StrEnum):
    SYS_MSG = "system_message"
    USR_MSG = "user_message"
    TOOL_MSG = "tool_message"
    TOOL_CALL = "tool_call"
    GEN_MSG = "gen_message"

    COMP = "completion"
    COMP_START = "completion_start"
    COMP_END = "completion_end"

    COMP_CHUNK = "completion_chunk"
    THINK_CHUNK = "thinking_chunk"
    RESP_CHUNK = "response_chunk"
    TOOL_CALL_CHUNK = "tool_call_chunk"
    ANNOT_CHUNK = "annotations_chunk"
    REFUSAL_CHUNK = "refusal_chunk"

    RESP_START = "response_start"
    RESP_END = "response_end"
    THINK_START = "thinking_start"
    THINK_END = "thinking_end"
    TOOL_CALL_START = "tool_call_start"
    TOOL_CALL_END = "tool_call_end"
    ANNOT_START = "annotations_start"
    ANNOT_END = "annotations_end"

    LLM_ERR = "llm_error"

    PACKET_OUT = "packet_output"
    PAYLOAD_OUT = "payload_output"
    PROC_ERR = "processor_error"
    PROC_START = "processor_start"
    PROC_END = "processor_end"

    WORKFLOW_RES = "workflow_result"
    RUN_RES = "run_result"


_T_co = TypeVar("_T_co", covariant=True)
_M_co = TypeVar("_M_co", covariant=True, bound=MessageBase)
_C_co = TypeVar("_C_co", covariant=True, bound=CompletionChunk)


class Event(BaseModel, Generic[_T_co], frozen=True):
    type: EventType
    source: EventSourceType
    id: str = Field(default_factory=lambda: str(uuid4()))
    created: int = Field(default_factory=lambda: int(time.time()))
    proc_name: str | None = None
    call_id: str | None = None
    data: _T_co


class DummyEvent(Event[Any], frozen=True):
    type: Literal[EventType.PAYLOAD_OUT] = EventType.PAYLOAD_OUT
    source: Literal[EventSourceType.PROC] = EventSourceType.PROC
    data: Any = None


# Non-streamed completion events


class CompletionEvent(Event[Completion], frozen=True):
    type: Literal[EventType.COMP] = EventType.COMP
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM


class MessageEvent(Event[_M_co], Generic[_M_co], frozen=True):
    pass


class GenMessageEvent(MessageEvent[AssistantMessage], frozen=True):
    type: Literal[EventType.GEN_MSG] = EventType.GEN_MSG
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM


class ToolMessageEvent(MessageEvent[ToolMessage], frozen=True):
    type: Literal[EventType.TOOL_MSG] = EventType.TOOL_MSG
    source: Literal[EventSourceType.TOOL] = EventSourceType.TOOL


class UserMessageEvent(MessageEvent[UserMessage], frozen=True):
    type: Literal[EventType.USR_MSG] = EventType.USR_MSG
    source: Literal[EventSourceType.USER] = EventSourceType.USER


class SystemMessageEvent(MessageEvent[SystemMessage], frozen=True):
    type: Literal[EventType.SYS_MSG] = EventType.SYS_MSG
    source: Literal[EventSourceType.AGENT] = EventSourceType.AGENT


class ToolCallEvent(Event[ToolCall], frozen=True):
    type: Literal[EventType.TOOL_CALL] = EventType.TOOL_CALL
    source: Literal[EventSourceType.AGENT] = EventSourceType.AGENT


# Streamed completion events

StreamedCompletionEventTypes = Literal[
    EventType.COMP_CHUNK,
    EventType.COMP_START,
    EventType.COMP_END,
    EventType.RESP_CHUNK,
    EventType.RESP_START,
    EventType.RESP_END,
    EventType.THINK_CHUNK,
    EventType.THINK_START,
    EventType.THINK_END,
    EventType.TOOL_CALL_CHUNK,
    EventType.TOOL_CALL_START,
    EventType.TOOL_CALL_END,
    EventType.ANNOT_CHUNK,
    EventType.ANNOT_START,
    EventType.ANNOT_END,
    EventType.REFUSAL_CHUNK,
]


class CompletionChunkEvent(Event[_C_co], Generic[_C_co], frozen=True):
    type: StreamedCompletionEventTypes = EventType.COMP_CHUNK
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM
    data: _C_co

    def split_into_specialized(
        self,
    ) -> "list[CompletionChunkEvent[Any]]":
        specialized_events: list[CompletionChunkEvent[Any]] = []

        for sub_chunk in self.data.split_into_specialized():
            if isinstance(sub_chunk, ResponseChunk):
                new_event = self.model_copy(
                    update={"data": sub_chunk, "type": EventType.RESP_CHUNK}
                )
                specialized_events.append(
                    ResponseChunkEvent.model_validate(new_event.model_dump())
                )
            if isinstance(sub_chunk, ThinkingChunk):
                new_event = self.model_copy(
                    update={"data": sub_chunk, "type": EventType.THINK_CHUNK}
                )
                specialized_events.append(
                    ThinkingChunkEvent.model_validate(new_event.model_dump())
                )
            if isinstance(sub_chunk, ToolCallChunk):
                new_event = self.model_copy(
                    update={"data": sub_chunk, "type": EventType.TOOL_CALL_CHUNK}
                )
                specialized_events.append(
                    ToolCallChunkEvent.model_validate(new_event.model_dump())
                )
            if isinstance(sub_chunk, AnnotationsChunk):
                new_event = self.model_copy(
                    update={"data": sub_chunk, "type": EventType.ANNOT_CHUNK}
                )
                specialized_events.append(
                    AnnotationsChunkEvent.model_validate(new_event.model_dump())
                )
            if isinstance(sub_chunk, RefusalChunk):
                new_event = self.model_copy(
                    update={"data": sub_chunk, "type": EventType.REFUSAL_CHUNK}
                )
                specialized_events.append(
                    RefusalChunkEvent.model_validate(new_event.model_dump())
                )

        return specialized_events


class ResponseChunkEvent(CompletionChunkEvent[ResponseChunk], frozen=True):
    type: Literal[EventType.RESP_CHUNK] = EventType.RESP_CHUNK
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM


class ThinkingChunkEvent(CompletionChunkEvent[ThinkingChunk], frozen=True):
    type: Literal[EventType.THINK_CHUNK] = EventType.THINK_CHUNK
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM


class ToolCallChunkEvent(CompletionChunkEvent[ToolCallChunk], frozen=True):
    type: Literal[EventType.TOOL_CALL_CHUNK] = EventType.TOOL_CALL_CHUNK
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM


class AnnotationsChunkEvent(CompletionChunkEvent[AnnotationsChunk], frozen=True):
    type: Literal[EventType.ANNOT_CHUNK] = EventType.ANNOT_CHUNK
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM


class RefusalChunkEvent(CompletionChunkEvent[RefusalChunk], frozen=True):
    type: Literal[EventType.REFUSAL_CHUNK] = EventType.REFUSAL_CHUNK
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM


START_END_MAP: dict[EventType, list[EventType]] = {
    EventType.COMP_CHUNK: [EventType.COMP_START, EventType.COMP_END],
    EventType.RESP_CHUNK: [EventType.RESP_START, EventType.RESP_END],
    EventType.THINK_CHUNK: [EventType.THINK_START, EventType.THINK_END],
    EventType.TOOL_CALL_CHUNK: [EventType.TOOL_CALL_START, EventType.TOOL_CALL_END],
    EventType.ANNOT_CHUNK: [EventType.ANNOT_START, EventType.ANNOT_END],
}


class LLMStateChangeEvent(CompletionChunkEvent[_C_co], Generic[_C_co], frozen=True):
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM
    start: bool = True

    @classmethod
    def from_chunk_event(
        cls, event: CompletionChunkEvent[CompletionChunk]
    ) -> "LLMStateChangeEvent[_C_co]":
        _type = get_args(cls.model_fields["type"].annotation)[0]
        return cls(**event.model_copy(update={"type": _type}).model_dump())


class CompletionStartEvent(LLMStateChangeEvent[CompletionChunk], frozen=True):
    type: Literal[EventType.COMP_START] = EventType.COMP_START
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM
    start: Literal[True] = True


class CompletionEndEvent(LLMStateChangeEvent[CompletionChunk], frozen=True):
    type: Literal[EventType.COMP_END] = EventType.COMP_END
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM
    start: Literal[False] = False


class ResponseStartEvent(LLMStateChangeEvent[ResponseChunk], frozen=True):
    type: Literal[EventType.RESP_START] = EventType.RESP_START
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM
    start: Literal[True] = True


class ResponseEndEvent(LLMStateChangeEvent[ResponseChunk], frozen=True):
    type: Literal[EventType.RESP_END] = EventType.RESP_END
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM
    start: Literal[False] = False


class ThinkingStartEvent(LLMStateChangeEvent[ThinkingChunk], frozen=True):
    type: Literal[EventType.THINK_START] = EventType.THINK_START
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM
    start: Literal[True] = True


class ThinkingEndEvent(LLMStateChangeEvent[ThinkingChunk], frozen=True):
    type: Literal[EventType.THINK_END] = EventType.THINK_END
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM
    start: Literal[False] = False


class ToolCallStartEvent(LLMStateChangeEvent[ToolCallChunk], frozen=True):
    type: Literal[EventType.TOOL_CALL_START] = EventType.TOOL_CALL_START
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM
    start: Literal[True] = True


class ToolCallEndEvent(LLMStateChangeEvent[ToolCallChunk], frozen=True):
    type: Literal[EventType.TOOL_CALL_END] = EventType.TOOL_CALL_END
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM
    start: Literal[False] = False


class AnnotationsStartEvent(LLMStateChangeEvent[AnnotationsChunk], frozen=True):
    type: Literal[EventType.ANNOT_START] = EventType.ANNOT_START
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM
    start: Literal[True] = True


class AnnotationsEndEvent(LLMStateChangeEvent[AnnotationsChunk], frozen=True):
    type: Literal[EventType.ANNOT_END] = EventType.ANNOT_END
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM
    start: Literal[False] = False


class LLMStreamingErrorData(BaseModel):
    error: Exception
    model_name: str | None = None
    model_id: str | None = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


class LLMStreamingErrorEvent(Event[LLMStreamingErrorData], frozen=True):
    type: Literal[EventType.LLM_ERR] = EventType.LLM_ERR
    source: Literal[EventSourceType.LLM] = EventSourceType.LLM


# Processor events


class ProcStartEvent(Event[None], frozen=True):
    type: Literal[EventType.PROC_START] = EventType.PROC_START
    source: Literal[EventSourceType.PROC] = EventSourceType.PROC


class ProcEndEvent(Event[None], frozen=True):
    type: Literal[EventType.PROC_END] = EventType.PROC_END
    source: Literal[EventSourceType.PROC] = EventSourceType.PROC


class ProcPayloadOutputEvent(Event[Any], frozen=True):
    type: Literal[EventType.PAYLOAD_OUT] = EventType.PAYLOAD_OUT
    source: Literal[EventSourceType.PROC] = EventSourceType.PROC


class ProcPacketOutputEvent(Event[Packet[Any]], frozen=True):
    type: Literal[EventType.PACKET_OUT, EventType.WORKFLOW_RES, EventType.RUN_RES] = (
        EventType.PACKET_OUT
    )
    source: Literal[
        EventSourceType.PROC, EventSourceType.WORKFLOW, EventSourceType.RUN
    ] = EventSourceType.PROC


class ProcStreamingErrorData(BaseModel):
    error: Exception
    call_id: str | None = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


class ProcStreamingErrorEvent(Event[ProcStreamingErrorData], frozen=True):
    type: Literal[EventType.PROC_ERR] = EventType.PROC_ERR
    source: Literal[EventSourceType.PROC] = EventSourceType.PROC


# Workflow and run events


class WorkflowResultEvent(ProcPacketOutputEvent, frozen=True):
    type: Literal[EventType.WORKFLOW_RES] = EventType.WORKFLOW_RES
    source: Literal[EventSourceType.WORKFLOW] = EventSourceType.WORKFLOW


class RunResultEvent(ProcPacketOutputEvent, frozen=True):
    type: Literal[EventType.RUN_RES] = EventType.RUN_RES
    source: Literal[EventSourceType.RUN] = EventSourceType.RUN



import os
import threading
from logging import getLogger

from opentelemetry.sdk.resources import SERVICE_NAME
from traceloop.sdk import Traceloop

from .exporters import NoopExporter

logger = getLogger(__name__)

_init_lock = threading.Lock()
_initialized = False


def init_traceloop(project_name: str = "grasp-agents"):
    global _initialized
    with _init_lock:
        if _initialized:
            return

        project_name_key = os.getenv("TELEMETRY_PROJECT_NAME_KEY", SERVICE_NAME)

        Traceloop.init(  # type: ignore
            app_name=project_name,
            exporter=NoopExporter(),
            resource_attributes={project_name_key: project_name},
            # disable_batch=False,
            # block_instruments={Instruments.OPENAI},
        )

        _initialized = True



import asyncio
from collections.abc import AsyncIterator, Coroutine
from datetime import UTC, datetime
from logging import getLogger
from pathlib import Path
from typing import Any, TypeVar

from tqdm.autonotebook import tqdm

logger = getLogger(__name__)


def read_txt(file_path: str | Path, encoding: str = "utf-8") -> str:
    return Path(file_path).read_text(encoding=encoding)


def read_contents_from_file(
    file_path: str | Path,
    binary_mode: bool = False,
) -> str | bytes:
    try:
        if binary_mode:
            return Path(file_path).read_bytes()
        return Path(file_path).read_text()
    except FileNotFoundError:
        logger.exception(f"File {file_path} not found.")
        return ""


def get_prompt(prompt_text: str | None, prompt_path: str | Path | None) -> str | None:
    if prompt_text is None:
        return read_contents_from_file(prompt_path) if prompt_path is not None else None  # type: ignore[arg-type]

    return prompt_text


async def asyncio_gather_with_pbar(
    *corouts: Coroutine[Any, Any, Any],
    no_tqdm: bool = False,
    desc: str | None = None,
) -> list[Any]:
    # TODO: optimize
    pbar = tqdm(total=len(corouts), desc=desc, disable=no_tqdm)

    async def run_and_update(coro: Coroutine[Any, Any, Any]) -> Any:
        result = await coro
        pbar.update(1)
        return result

    wrapped_tasks = [run_and_update(c) for c in corouts]
    results = await asyncio.gather(*wrapped_tasks)
    pbar.close()

    return results


def get_timestamp() -> str:
    return datetime.now(UTC).strftime("%Y%m%d_%H%M%S")


_T = TypeVar("_T")


async def stream_concurrent(
    generators: list[AsyncIterator[_T]],
) -> AsyncIterator[tuple[int, _T]]:
    queue: asyncio.Queue[tuple[int, _T] | None] = asyncio.Queue()
    pumps_left = len(generators)

    async def pump(gen: AsyncIterator[_T], idx: int) -> None:
        nonlocal pumps_left
        try:
            async for item in gen:
                await queue.put((idx, item))
        finally:
            pumps_left -= 1
            if pumps_left == 0:
                await queue.put(None)

    for idx, gen in enumerate(generators):
        asyncio.create_task(pump(gen, idx))

    while True:
        msg = await queue.get()
        if msg is None:
            break
        yield msg


def is_method_overridden(
    method_name: str, self: object, base_cls: type[Any] | None = None
) -> bool:
    """
    Check if a method is overridden in a subclass compared to a base class
    or if it is defined directly in the instance's __dict__ (e.g. via a decorator).
    """
    child_cls = type(self)
    if not hasattr(child_cls, method_name):
        raise AttributeError(f"{child_cls} has no method named {method_name}")

    set_via_decorator = method_name in self.__dict__

    overriden_in_child = False
    if base_cls is not None:
        overriden_in_child = hasattr(base_cls, method_name) and getattr(
            child_cls, method_name
        ) is not getattr(base_cls, method_name)

    return set_via_decorator or overriden_in_child



import asyncio
import json
from collections.abc import AsyncIterator, Coroutine, Mapping, Sequence
from copy import deepcopy
from itertools import starmap
from logging import getLogger
from typing import Any, Generic, Protocol, TypedDict, final

from pydantic import BaseModel

from grasp_agents.tracing_decorators import task
from grasp_agents.typing.completion_chunk import CompletionChunk

from .errors import AgentFinalAnswerError
from .llm import LLM
from .llm_agent_memory import LLMAgentMemory
from .run_context import CtxT, RunContext
from .typing.completion import Completion
from .typing.events import (
    CompletionChunkEvent,
    CompletionEvent,
    Event,
    GenMessageEvent,
    LLMStreamingErrorEvent,
    ToolCallEvent,
    ToolMessageEvent,
    UserMessageEvent,
)
from .typing.message import AssistantMessage, ToolMessage, UserMessage
from .typing.tool import BaseTool, NamedToolChoice, ToolCall, ToolChoice
from .utils import is_method_overridden

logger = getLogger(__name__)


class FinalAnswerChecker(Protocol[CtxT]):
    def __call__(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> str | None: ...


class BeforeGenerateHook(Protocol[CtxT]):
    async def __call__(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
        llm_settings: dict[str, Any],
    ) -> None: ...


class AfterGenerateHook(Protocol[CtxT]):
    async def __call__(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
        llm_settings: dict[str, Any],
    ) -> None: ...


class ToolOutputConverter(Protocol[CtxT]):
    def __call__(
        self,
        tool_outputs: Sequence[Any],
        tool_calls: Sequence[ToolCall],
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> Sequence[ToolMessage | UserMessage]: ...


class HookArgs(TypedDict, total=False):
    memory: LLMAgentMemory
    ctx: RunContext[Any]
    call_id: str
    llm_settings: dict[str, Any]


class LLMPolicyExecutor(Generic[CtxT]):
    def __init__(
        self,
        *,
        agent_name: str,
        llm: LLM,
        tools: list[BaseTool[BaseModel, Any, CtxT]] | None,
        response_schema: Any | None = None,
        response_schema_by_xml_tag: Mapping[str, Any] | None = None,
        max_turns: int,
        react_mode: bool = False,
        final_answer_type: type[BaseModel] = BaseModel,
        final_answer_as_tool_call: bool = False,
    ) -> None:
        super().__init__()

        self._agent_name = agent_name
        self._max_turns = max_turns
        self._react_mode = react_mode

        self._llm = llm
        self._response_schema = response_schema
        self._response_schema_by_xml_tag = response_schema_by_xml_tag

        self._final_answer_type = final_answer_type
        self._final_answer_as_tool_call = final_answer_as_tool_call
        self._final_answer_tool = self.get_final_answer_tool()

        tools_list: list[BaseTool[BaseModel, Any, CtxT]] | None = tools
        if tools and final_answer_as_tool_call:
            tools_list = tools + [self._final_answer_tool]
        self._tools = {t.name: t for t in tools_list} if tools_list else None

    @property
    def agent_name(self) -> str:
        return self._agent_name

    @property
    def max_turns(self) -> int:
        return self._max_turns

    @property
    def react_mode(self) -> bool:
        return self._react_mode

    @property
    def llm(self) -> LLM:
        return self._llm

    @property
    def response_schema(self) -> Any | None:
        return self._response_schema

    @response_schema.setter
    def response_schema(self, value: Any | None) -> None:
        self._response_schema = value

    @property
    def response_schema_by_xml_tag(self) -> Mapping[str, Any] | None:
        return self._response_schema_by_xml_tag

    @property
    def tools(self) -> dict[str, BaseTool[BaseModel, Any, CtxT]]:
        return self._tools or {}

    def check_for_final_answer_impl(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> str | None:
        raise NotImplementedError

    @final
    def check_for_final_answer(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> str | None:
        if is_method_overridden("check_for_final_answer_impl", self):
            return self.check_for_final_answer_impl(
                memory, ctx=ctx, call_id=call_id, **kwargs
            )

        if self._final_answer_as_tool_call:
            return self._get_final_answer_from_tool_call(memory)

        return None

    async def on_before_generate_impl(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
        llm_settings: dict[str, Any],
    ) -> None:
        raise NotImplementedError

    @final
    async def on_before_generate(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
        llm_settings: dict[str, Any],
    ) -> None:
        if is_method_overridden("on_before_generate_impl", self):
            await self.on_before_generate_impl(
                memory=memory,
                ctx=ctx,
                call_id=call_id,
                num_turns=num_turns,
                llm_settings=llm_settings,
            )

    async def on_after_generate_impl(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
        llm_settings: dict[str, Any],
    ) -> None:
        raise NotImplementedError

    @final
    async def on_after_generate(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
        llm_settings: dict[str, Any],
    ) -> None:
        if is_method_overridden("on_after_generate_impl", self):
            await self.on_after_generate_impl(
                memory=memory,
                ctx=ctx,
                call_id=call_id,
                num_turns=num_turns,
                llm_settings=llm_settings,
            )

    @task(name="generate")  # type: ignore
    async def generate_message(
        self,
        memory: LLMAgentMemory,
        *,
        tool_choice: ToolChoice | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
        extra_llm_settings: dict[str, Any],
    ) -> AssistantMessage:
        completion = await self.llm.generate_completion(
            memory.messages,
            response_schema=self.response_schema,
            response_schema_by_xml_tag=self.response_schema_by_xml_tag,
            tools=self.tools,
            tool_choice=tool_choice,
            proc_name=self.agent_name,
            call_id=call_id,
            **extra_llm_settings,
        )
        memory.update([completion.message])
        self._process_completion(completion, ctx=ctx, call_id=call_id)

        return completion.message

    @task(name="generate")  # type: ignore
    async def generate_message_stream(
        self,
        memory: LLMAgentMemory,
        *,
        tool_choice: ToolChoice | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
        extra_llm_settings: dict[str, Any],
    ) -> AsyncIterator[
        CompletionChunkEvent[CompletionChunk]
        | CompletionEvent
        | GenMessageEvent
        | LLMStreamingErrorEvent
    ]:
        completion: Completion | None = None

        llm_event_stream = self.llm.generate_completion_stream(
            memory.messages,
            response_schema=self.response_schema,
            response_schema_by_xml_tag=self.response_schema_by_xml_tag,
            tools=self.tools,
            tool_choice=tool_choice,
            proc_name=self.agent_name,
            call_id=call_id,
            **extra_llm_settings,
        )
        llm_event_stream_post = self.llm.postprocess_event_stream(llm_event_stream)

        async for event in llm_event_stream_post:
            if isinstance(event, CompletionEvent):
                completion = event.data
            yield event
        if completion is None:
            return
        yield GenMessageEvent(
            proc_name=self.agent_name, call_id=call_id, data=completion.message
        )

        memory.update([completion.message])
        self._process_completion(completion, ctx=ctx, call_id=call_id)

    def tool_outputs_to_messages_impl(
        self,
        tool_outputs: Sequence[Any],
        tool_calls: Sequence[ToolCall],
        *,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> Sequence[ToolMessage | UserMessage]:
        raise NotImplementedError

    def tool_outputs_to_messages_default(
        self, tool_outputs: Sequence[Any], tool_calls: Sequence[ToolCall]
    ) -> Sequence[ToolMessage | UserMessage]:
        return list(
            starmap(
                ToolMessage.from_tool_output, zip(tool_outputs, tool_calls, strict=True)
            )
        )

    @final
    def tool_outputs_to_messages(
        self,
        tool_outputs: Sequence[Any],
        tool_calls: Sequence[ToolCall],
        *,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> Sequence[ToolMessage | UserMessage]:
        if is_method_overridden("tool_outputs_to_messages_impl", self):
            return self.tool_outputs_to_messages_impl(
                tool_outputs, tool_calls, ctx=ctx, call_id=call_id
            )
        return self.tool_outputs_to_messages_default(tool_outputs, tool_calls)

    # @task(name="call_tools")  # type: ignore
    async def call_tools(
        self,
        calls: Sequence[ToolCall],
        memory: LLMAgentMemory,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> Sequence[ToolMessage | UserMessage]:
        corouts: list[Coroutine[Any, Any, BaseModel]] = []
        for call in calls:
            tool = self.tools[call.tool_name]
            args = json.loads(call.tool_arguments)
            corouts.append(tool(ctx=ctx, call_id=call_id, **args))

        outs = await asyncio.gather(*corouts)
        tool_messages = self.tool_outputs_to_messages(
            outs, calls, ctx=ctx, call_id=call_id
        )
        memory.update(tool_messages)

        if ctx.printer:
            ctx.printer.print_messages(
                tool_messages, agent_name=self.agent_name, call_id=call_id
            )

        return tool_messages

    # @task(name="call_tools")  # type: ignore
    async def call_tools_stream(
        self,
        calls: Sequence[ToolCall],
        memory: LLMAgentMemory,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> AsyncIterator[ToolCallEvent | ToolMessageEvent | UserMessageEvent]:
        for tool_call in calls:
            yield ToolCallEvent(
                proc_name=self.agent_name, call_id=call_id, data=tool_call
            )

        tool_messages = await self.call_tools(
            calls, memory=memory, ctx=ctx, call_id=call_id
        )

        for tool_message, call in zip(tool_messages, calls, strict=True):
            if isinstance(tool_message, UserMessage):
                yield UserMessageEvent(
                    proc_name=call.tool_name, call_id=call_id, data=tool_message
                )
            else:
                yield ToolMessageEvent(
                    proc_name=call.tool_name, call_id=call_id, data=tool_message
                )

    def _get_final_answer_from_tool_call(self, memory: LLMAgentMemory) -> str | None:
        msgs = memory.messages
        if (
            msgs
            and isinstance(msgs[-1], AssistantMessage)
            and msgs[-1].tool_calls
            and msgs[-1].tool_calls[0].tool_name == self._final_answer_tool.name
        ):
            return msgs[-1].tool_calls[0].tool_arguments
        return None

    def _get_final_answer_from_message(self, memory: LLMAgentMemory) -> str | None:
        msgs = memory.messages
        if msgs and isinstance(msgs[-1], AssistantMessage) and msgs[-1].content:
            return msgs[-1].content
        return None

    def get_final_answer(self, memory: LLMAgentMemory) -> str | None:
        if self._final_answer_as_tool_call:
            return self._get_final_answer_from_tool_call(memory)
        return self._get_final_answer_from_message(memory)

    @task(name="force_generate_final_answer")  # type: ignore
    async def _force_generate_final_answer(
        self,
        memory: LLMAgentMemory,
        ctx: RunContext[CtxT],
        call_id: str,
        extra_llm_settings: dict[str, Any],
    ) -> str:
        # NOTE: Might not need the user message when forcing the tool call
        user_message = UserMessage.from_text(
            "Exceeded the maximum number of turns: provide a final answer now!"
        )
        memory.update([user_message])

        if ctx.printer:
            ctx.printer.print_messages(
                [user_message], agent_name=self.agent_name, call_id=call_id
            )

        tool_choice = (
            NamedToolChoice(name=self._final_answer_tool.name)
            if self._final_answer_as_tool_call
            else None
        )
        _ = await self.generate_message(
            memory,
            tool_choice=tool_choice,
            ctx=ctx,
            call_id=call_id,
            extra_llm_settings=extra_llm_settings,
        )

        final_answer = self.get_final_answer(memory)
        if final_answer is None:
            raise AgentFinalAnswerError(proc_name=self.agent_name, call_id=call_id)

        return final_answer

    @task(name="force_generate_final_answer")  # type: ignore
    async def _force_generate_final_answer_stream(
        self,
        memory: LLMAgentMemory,
        ctx: RunContext[CtxT],
        call_id: str,
        extra_llm_settings: dict[str, Any],
    ) -> AsyncIterator[Event[Any]]:
        # NOTE: Might not need the user message when forcing the tool call
        user_message = UserMessage.from_text(
            "Exceeded the maximum number of turns: provide a final answer now!",
        )
        memory.update([user_message])
        yield UserMessageEvent(
            proc_name=self.agent_name, call_id=call_id, data=user_message
        )
        if ctx.printer:
            ctx.printer.print_messages(
                [user_message], agent_name=self.agent_name, call_id=call_id
            )

        tool_choice = (
            NamedToolChoice(name=self._final_answer_tool.name)
            if self._final_answer_as_tool_call
            else None
        )
        async for event in self.generate_message_stream(
            memory,
            tool_choice=tool_choice,
            ctx=ctx,
            call_id=call_id,
            extra_llm_settings=extra_llm_settings,
        ):
            yield event

        final_answer = self.get_final_answer(memory)
        if final_answer is None:
            raise AgentFinalAnswerError(proc_name=self.agent_name, call_id=call_id)

    async def execute(
        self,
        memory: LLMAgentMemory,
        ctx: RunContext[CtxT],
        call_id: str,
        extra_llm_settings: dict[str, Any] | None = None,
    ) -> str:
        """
        Some LLMs do not output tool calls and message content in the same response.
        To enable planning/observation before/after tool calls for such models,
        we might want to force the agent to output a message without
        tool calls (planning) first, then force tool calls in the next message, etc.
        For this, we use the `react_mode` flag.
        """
        _extra_llm_settings = deepcopy(extra_llm_settings or {})
        hooks_kwargs: HookArgs = HookArgs(
            memory=memory,
            ctx=ctx,
            call_id=call_id,
            llm_settings=_extra_llm_settings,
        )

        turns = 0

        # 1. Generate the first message and update memory

        # In ReAct mode, we generate the first message without tool calls
        # to enforce planning.

        tool_choice: ToolChoice | None = None
        if self.tools:
            tool_choice = "none" if self.react_mode else "auto"

        # LLM settings can be modified in-place
        await self.on_before_generate(**hooks_kwargs, num_turns=turns)

        gen_message = await self.generate_message(
            memory,
            tool_choice=tool_choice,
            ctx=ctx,
            call_id=call_id,
            extra_llm_settings=_extra_llm_settings,
        )

        # LLM settings can be modified in-place
        await self.on_after_generate(**hooks_kwargs, num_turns=turns)

        if not self.tools:
            return gen_message.content or ""

        while True:
            # 2. Check if we have a final answer

            final_answer = self.check_for_final_answer(
                memory, ctx=ctx, call_id=call_id, num_turns=turns
            )
            if final_answer is not None:
                return final_answer

            if turns >= self.max_turns:
                final_answer = await self._force_generate_final_answer(
                    memory,
                    ctx=ctx,
                    call_id=call_id,
                    extra_llm_settings=_extra_llm_settings,
                )
                logger.info(
                    f"Max turns reached: {self.max_turns}. Exiting the tool call loop."
                )
                return final_answer

            # 3. Call tools and update memory

            if gen_message.tool_calls:
                await self.call_tools(
                    gen_message.tool_calls, memory=memory, ctx=ctx, call_id=call_id
                )

            # 4. Generate the next message and update memory

            if self.react_mode and gen_message.tool_calls:
                # ReAct mode: used tools in the last message -> avoid tool calls in the next message
                tool_choice = "none"
            elif self.react_mode:
                # ReAct mode: no tool calls in the last message -> force tool calls in the next message
                tool_choice = "required"
            else:
                # No ReAct mode: let the model decide
                tool_choice = "auto"

            await self.on_before_generate(**hooks_kwargs, num_turns=turns)

            gen_message = await self.generate_message(
                memory,
                tool_choice=tool_choice,
                ctx=ctx,
                call_id=call_id,
                extra_llm_settings=_extra_llm_settings,
            )

            await self.on_after_generate(**hooks_kwargs, num_turns=turns)

            turns += 1

    async def execute_stream(
        self,
        memory: LLMAgentMemory,
        ctx: RunContext[CtxT],
        call_id: str,
        extra_llm_settings: dict[str, Any] | None = None,
    ) -> AsyncIterator[Event[Any]]:
        _extra_llm_settings = deepcopy(extra_llm_settings or {})
        hooks_kwargs: HookArgs = HookArgs(
            memory=memory,
            ctx=ctx,
            call_id=call_id,
            llm_settings=_extra_llm_settings,
        )

        turns = 0

        # 1. Generate the first message and update memory

        tool_choice: ToolChoice | None = None
        if self.tools:
            tool_choice = "none" if self.react_mode else "auto"

        await self.on_before_generate(**hooks_kwargs, num_turns=turns)

        gen_message: AssistantMessage | None = None
        async for event in self.generate_message_stream(
            memory,
            tool_choice=tool_choice,
            ctx=ctx,
            call_id=call_id,
            extra_llm_settings=_extra_llm_settings,
        ):
            if isinstance(event, GenMessageEvent):
                gen_message = event.data
            yield event

        if gen_message is None:
            # Exit if generation failed
            return

        await self.on_after_generate(**hooks_kwargs, num_turns=turns)

        if not self.tools:
            # No tools to call, return the content of the generated message
            return

        while True:
            # 2. Check if we have a final answer

            final_answer = self.check_for_final_answer(
                memory, ctx=ctx, call_id=call_id, num_turns=turns
            )
            if final_answer is not None:
                return

            if turns >= self.max_turns:
                async for event in self._force_generate_final_answer_stream(
                    memory,
                    ctx=ctx,
                    call_id=call_id,
                    extra_llm_settings=_extra_llm_settings,
                ):
                    yield event
                logger.info(
                    f"Max turns reached: {self.max_turns}. Exiting the tool call loop."
                )
                return

            # 3. Call tools and update memory

            if gen_message.tool_calls:
                async for event in self.call_tools_stream(
                    gen_message.tool_calls, memory=memory, ctx=ctx, call_id=call_id
                ):
                    yield event

            # 4. Generate the next message and update memory

            if self.react_mode and gen_message.tool_calls:
                tool_choice = "none"
            elif self.react_mode:
                tool_choice = "required"
            else:
                tool_choice = "auto"

            await self.on_before_generate(**hooks_kwargs, num_turns=turns)

            async for event in self.generate_message_stream(
                memory,
                tool_choice=tool_choice,
                ctx=ctx,
                call_id=call_id,
                extra_llm_settings=_extra_llm_settings,
            ):
                yield event
                if isinstance(event, GenMessageEvent):
                    gen_message = event.data

            await self.on_after_generate(**hooks_kwargs, num_turns=turns)

            turns += 1

    def get_final_answer_tool(self) -> BaseTool[BaseModel, None, Any]:
        class FinalAnswerTool(BaseTool[self._final_answer_type, None, Any]):
            name: str = "final_answer"
            description: str = (
                "You must call this tool to provide the final answer. "
                "DO NOT output your answer before calling the tool. "
            )

            async def run(
                self,
                inp: BaseModel,
                *,
                ctx: RunContext[Any] | None = None,
                call_id: str | None = None,
            ) -> None:
                return None

        return FinalAnswerTool()

    def _process_completion(
        self, completion: Completion, *, ctx: RunContext[CtxT], call_id: str
    ) -> None:
        ctx.completions[self.agent_name].append(completion)
        ctx.usage_tracker.update(
            agent_name=self.agent_name,
            completions=[completion],
            model_name=self.llm.model_name,
        )
        if ctx.printer:
            ctx.printer.print_messages(
                [completion.message],
                usages=[completion.usage],
                agent_name=self.agent_name,
                call_id=call_id,
            )



import asyncio
import logging
from collections.abc import AsyncIterator
from types import TracebackType
from typing import Any, Literal, Protocol, TypeVar

from .packet import Packet
from .typing.events import Event
from .typing.io import ProcName

logger = logging.getLogger(__name__)


START_PROC_NAME: Literal["*START*"] = "*START*"
END_PROC_NAME: Literal["*END*"] = "*END*"


_PayloadT_contra = TypeVar("_PayloadT_contra", contravariant=True)


class PacketHandler(Protocol[_PayloadT_contra]):
    async def __call__(
        self, packet: Packet[_PayloadT_contra], **kwargs: Any
    ) -> None: ...


class PacketPool:
    def __init__(self) -> None:
        self._packet_queues: dict[ProcName, asyncio.Queue[Packet[Any] | None]] = {}
        self._packet_handlers: dict[ProcName, PacketHandler[Any]] = {}
        self._task_group: asyncio.TaskGroup | None = None

        self._event_queue: asyncio.Queue[Event[Any] | None] = asyncio.Queue()

        self._final_result_fut: asyncio.Future[Packet[Any]]

        self._stopping = False
        self._stopped_evt = asyncio.Event()

        self._errors: list[Exception] = []

    async def post(self, packet: Packet[Any]) -> None:
        if packet.uniform_routing == [END_PROC_NAME]:
            if not self._final_result_fut.done():
                self._final_result_fut.set_result(packet)
            await self.shutdown()
            return

        for sub_packet in packet.split_by_recipient() or []:
            if not sub_packet.routing:
                continue
            if not sub_packet.routing[0]:
                continue
            recipient = sub_packet.routing[0][0]
            queue = self._packet_queues.setdefault(recipient, asyncio.Queue())
            await queue.put(sub_packet)

    async def final_result(self) -> Packet[Any]:
        try:
            return await self._final_result_fut
        finally:
            await self.shutdown()

    def register_packet_handler(
        self, proc_name: ProcName, handler: PacketHandler[Any]
    ) -> None:
        if self._stopping:
            raise RuntimeError("PacketPool is stopping/stopped")

        self._packet_handlers[proc_name] = handler
        self._packet_queues.setdefault(proc_name, asyncio.Queue())

        if self._task_group is not None:
            self._task_group.create_task(
                self._handle_packets(proc_name),
                name=f"packet-handler:{proc_name}",
            )

    async def push_event(self, event: Event[Any]) -> None:
        await self._event_queue.put(event)

    async def __aenter__(self) -> "PacketPool":
        self._task_group = asyncio.TaskGroup()
        await self._task_group.__aenter__()

        self._final_result_fut = asyncio.get_running_loop().create_future()

        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> bool | None:
        await self.shutdown()

        if self._task_group is not None:
            try:
                return await self._task_group.__aexit__(exc_type, exc, tb)
            finally:
                self._task_group = None

        if self._errors:
            raise ExceptionGroup("PacketPool worker errors", self._errors)

        return False

    async def _handle_packets(self, proc_name: ProcName) -> None:
        queue = self._packet_queues[proc_name]
        handler = self._packet_handlers[proc_name]

        while True:
            packet = await queue.get()
            if packet is None:
                break

            if self._final_result_fut.done():
                continue

            try:
                await handler(packet)
            except asyncio.CancelledError:
                raise
            except Exception as err:
                logger.exception("Error handling packet for %s", proc_name)
                self._errors.append(err)
                if not self._final_result_fut.done():
                    self._final_result_fut.set_exception(err)
                await self.shutdown()
                raise

    async def stream_events(self) -> AsyncIterator[Event[Any]]:
        while True:
            event = await self._event_queue.get()
            if event is None:
                break
            yield event

    async def shutdown(self) -> None:
        if self._stopping:
            await self._stopped_evt.wait()
            return
        self._stopping = True
        try:
            await self._event_queue.put(None)
            for queue in self._packet_queues.values():
                await queue.put(None)
        finally:
            self._stopped_evt.set()



from collections import defaultdict
from typing import Generic, TypeVar

from pydantic import BaseModel, ConfigDict, Field

from grasp_agents.typing.completion import Completion

from .printer import Printer
from .typing.io import ProcName
from .usage_tracker import UsageTracker

CtxT = TypeVar("CtxT")


class RunContext(BaseModel, Generic[CtxT]):
    state: CtxT = None  # type: ignore

    completions: dict[ProcName, list[Completion]] = Field(
        default_factory=lambda: defaultdict(list)
    )
    usage_tracker: UsageTracker = Field(default_factory=UsageTracker, exclude=True)
    printer: Printer | None = Field(default=None, exclude=True)

    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)



from grasp_agents.errors import CompletionError
from grasp_agents.typing.completion import Usage

from ..openai.completion_converters import from_api_completion_usage
from ..typing.completion_chunk import (
    CompletionChunk,
    CompletionChunkDelta,
    CompletionChunkDeltaToolCall,
)
from . import LiteLLMCompletionChunk


def from_api_completion_chunk(
    api_completion_chunk: LiteLLMCompletionChunk, name: str | None = None
) -> CompletionChunk:
    if len(api_completion_chunk.choices) > 1:
        raise CompletionError("Multiple choices are not supported")
    api_choice = api_completion_chunk.choices[0]

    api_delta = api_choice.delta

    delta = CompletionChunkDelta(
        tool_calls=[
            CompletionChunkDeltaToolCall(
                id=tool_call.id,
                index=tool_call.index,
                tool_name=tool_call.function.name,
                tool_arguments=tool_call.function.arguments,
            )
            for tool_call in (api_delta.tool_calls or [])
            if tool_call.function
        ],
        content=api_delta.content,  # type: ignore[assignment, arg-type]
        role=api_delta.role,  # type: ignore[assignment, arg-type]
        thinking_blocks=getattr(api_delta, "thinking_blocks", None),
        annotations=getattr(api_delta, "annotations", None),
        reasoning_content=getattr(api_delta, "reasoning_content", None),
        provider_specific_fields=api_delta.provider_specific_fields,
        refusal=getattr(api_delta, "refusal", None),
    )

    usage: Usage | None = None
    api_usage = getattr(api_completion_chunk, "usage", None)
    if api_usage is not None:
        usage = from_api_completion_usage(api_usage)
        hidden_params = getattr(api_completion_chunk, "_hidden_params", {})
        usage.cost = getattr(hidden_params, "response_cost", None)

    return CompletionChunk(
        id=api_completion_chunk.id,
        model=api_completion_chunk.model,
        name=name,
        created=api_completion_chunk.created,
        system_fingerprint=api_completion_chunk.system_fingerprint,
        delta=delta,
        finish_reason=api_choice.finish_reason,  # type: ignore[assignment, arg-type]
        logprobs=getattr(api_choice, "logprobs", None),
        usage=usage,
        provider_specific_fields=api_completion_chunk.provider_specific_fields,
        hidden_params=api_completion_chunk._hidden_params,  # type: ignore[union-attr]
        response_ms=getattr(api_completion_chunk, "_response_ms", None),
    )



from ..typing.message import (
    AssistantMessage,
)
from ..typing.tool import ToolCall
from . import LiteLLMCompletionMessage, LiteLLMFunction, LiteLLMToolCall


def from_api_assistant_message(
    api_message: LiteLLMCompletionMessage, name: str | None = None
) -> AssistantMessage:
    tool_calls = None
    if api_message.tool_calls is not None:
        tool_calls = [
            ToolCall(
                id=tool_call.id,
                tool_name=tool_call.function.name,  # type: ignore
                tool_arguments=tool_call.function.arguments,
            )
            for tool_call in api_message.tool_calls
        ]

    return AssistantMessage(
        content=api_message.content,
        tool_calls=tool_calls,
        name=name,
        thinking_blocks=getattr(api_message, "thinking_blocks", None),
        reasoning_content=getattr(api_message, "reasoning_content", None),
        annotations=getattr(api_message, "annotations", None),
        provider_specific_fields=api_message.provider_specific_fields,
        refusal=getattr(api_message, "refusal", None),
    )


def to_api_assistant_message(
    message: AssistantMessage,
) -> LiteLLMCompletionMessage:
    api_tool_calls = None
    if message.tool_calls is not None:
        api_tool_calls = [
            LiteLLMToolCall(
                type="function",
                id=tool_call.id,
                function=LiteLLMFunction(
                    name=tool_call.tool_name,
                    arguments=tool_call.tool_arguments,
                ),
            )
            for tool_call in message.tool_calls
        ]

    api_message = LiteLLMCompletionMessage(role="assistant", content=message.content)

    if api_tool_calls:
        api_message.tool_calls = api_tool_calls

    for key in [
        "thinking_blocks",
        "reasoning_content",
        "annotations",
        "provider_specific_fields",
        "refusal",
    ]:
        if getattr(message, key):
            api_message[key] = getattr(message, key)

    return api_message



import base64
import re
from collections.abc import Iterable
from enum import StrEnum
from pathlib import Path
from typing import Annotated, Any, Literal, TypeAlias

from pydantic import AnyUrl, BaseModel, Field


class ContentType(StrEnum):
    TEXT = "text"
    IMAGE = "image"


ImageDetail: TypeAlias = Literal["low", "high", "auto"]


class ImageData(BaseModel):
    type: Literal["url", "base64"]
    url: AnyUrl | None = None
    base64: str | None = None

    # Supported by OpenAI API
    detail: ImageDetail = "high"

    @classmethod
    def from_base64(cls, base64_encoding: str, **kwargs: Any) -> "ImageData":
        return cls(type="base64", base64=base64_encoding, **kwargs)

    @classmethod
    def from_path(cls, img_path: str | Path, **kwargs: Any) -> "ImageData":
        image_bytes = Path(img_path).read_bytes()
        base64_encoding = base64.b64encode(image_bytes).decode("utf-8")
        return cls(type="base64", base64=base64_encoding, **kwargs)

    @classmethod
    def from_url(cls, img_url: str, **kwargs: Any) -> "ImageData":
        return cls(type="url", url=img_url, **kwargs)  # type: ignore

    def to_str(self) -> str:
        if self.type == "url":
            return str(self.url)
        if self.type == "base64":
            return str(self.base64)
        raise ValueError(f"Unsupported image data type: {self.type}")


class ContentPartText(BaseModel):
    type: Literal[ContentType.TEXT] = ContentType.TEXT
    data: str


class ContentPartImage(BaseModel):
    type: Literal[ContentType.IMAGE] = ContentType.IMAGE
    data: ImageData


ContentPart = Annotated[ContentPartText | ContentPartImage, Field(discriminator="type")]


class Content(BaseModel):
    parts: list[ContentPart]

    @classmethod
    def from_formatted_prompt(
        cls,
        prompt_template: str,
        /,
        **prompt_args: str | int | bool | ImageData | None,
    ) -> "Content":
        prompt_args = prompt_args or {}
        image_args = {
            arg_name: arg_val
            for arg_name, arg_val in prompt_args.items()
            if isinstance(arg_val, ImageData)
        }
        text_args = {
            arg_name: arg_val
            for arg_name, arg_val in prompt_args.items()
            if isinstance(arg_val, (str, int, float))
        }

        if not image_args:
            prompt_with_args = prompt_template.format(**text_args)
            return cls(parts=[ContentPartText(data=prompt_with_args)])

        pattern = r"({})".format("|".join([r"\{" + s + r"\}" for s in image_args]))
        input_prompt_chunks = re.split(pattern, prompt_template)

        content_parts: list[ContentPart] = []
        for chunk in input_prompt_chunks:
            stripped_chunk = chunk.strip(" \n")
            if re.match(pattern, stripped_chunk):
                image_data = image_args[stripped_chunk[1:-1]]
                content_part = ContentPartImage(data=image_data)
            else:
                text_data = stripped_chunk.format(**text_args)
                content_part = ContentPartText(data=text_data)
            content_parts.append(content_part)

        return cls(parts=content_parts)

    @classmethod
    def from_text(cls, text: str) -> "Content":
        return cls(parts=[ContentPartText(data=text)])

    @classmethod
    def from_image(cls, image: ImageData) -> "Content":
        return cls(parts=[ContentPartImage(data=image)])

    @classmethod
    def from_images(cls, images: Iterable[ImageData]) -> "Content":
        return cls(parts=[ContentPartImage(data=image) for image in images])

    @classmethod
    def from_content_parts(cls, content_parts: Iterable[str | ImageData]) -> "Content":
        parts: list[ContentPart] = []
        for part in content_parts:
            if isinstance(part, str):
                parts.append(ContentPartText(data=part))
            else:
                parts.append(ContentPartImage(data=part))

        return cls(parts=parts)



from collections.abc import Callable, Coroutine
from typing import Any, Concatenate, ParamSpec, TypeAlias, TypeVar

R = TypeVar("R")
P = ParamSpec("P")

AsyncCallable: TypeAlias = Callable[P, Coroutine[Any, Any, R]]
AsyncFunction: TypeAlias = Callable[P, Coroutine[Any, Any, R]]
AsyncMethod: TypeAlias = Callable[Concatenate[Any, P], Coroutine[Any, Any, R]]
AsyncFunctionOrMethod: TypeAlias = AsyncFunction[P, R] | AsyncMethod[P, R]



import warnings
from collections.abc import Awaitable, Callable
from typing import Any, Optional, ParamSpec, TypeVar

from opentelemetry.semconv_ai import TraceloopSpanKindValues

from .base import entity_class, entity_method

P = ParamSpec("P")
R = TypeVar("R")
F = TypeVar("F", bound=Callable[P, R | Awaitable[R]])


def task(
    name: str | None = None,
    version: int | None = None,
    method_name: str | None = None,
    tlp_span_kind: TraceloopSpanKindValues | None = TraceloopSpanKindValues.TASK,
) -> Callable[[F], F]:
    if method_name is None:
        return entity_method(name=name, version=version, tlp_span_kind=tlp_span_kind)
    return entity_class(
        name=name,
        version=version,
        method_name=method_name,
        tlp_span_kind=tlp_span_kind,
    )


def workflow(
    name: str | None = None,
    version: int | None = None,
    method_name: str | None = None,
    tlp_span_kind: TraceloopSpanKindValues | None = TraceloopSpanKindValues.WORKFLOW,
) -> Callable[[F], F]:
    if method_name is None:
        return entity_method(name=name, version=version, tlp_span_kind=tlp_span_kind)
    return entity_class(
        name=name,
        version=version,
        method_name=method_name,
        tlp_span_kind=tlp_span_kind,
    )


def agent(
    name: str | None = None,
    version: int | None = None,
    method_name: str | None = None,
) -> Callable[[F], F]:
    return workflow(
        name=name,
        version=version,
        method_name=method_name,
        tlp_span_kind=TraceloopSpanKindValues.AGENT,
    )


def tool(
    name: str | None = None,
    version: int | None = None,
    method_name: str | None = None,
) -> Callable[[F], F]:
    return task(
        name=name,
        version=version,
        method_name=method_name,
        tlp_span_kind=TraceloopSpanKindValues.TOOL,
    )



import logging
from abc import abstractmethod
from collections.abc import AsyncIterator, Mapping
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Required

import httpx
from pydantic import BaseModel
from typing_extensions import TypedDict

from .llm import LLM, LLMSettings
from .rate_limiting.rate_limiter import RateLimiter, limit_rate
from .typing.completion import Completion
from .typing.completion_chunk import CompletionChunk
from .typing.events import (
    CompletionChunkEvent,
    CompletionEvent,
    LLMStreamingErrorEvent,
)
from .typing.message import AssistantMessage, Messages
from .typing.tool import BaseTool, ToolChoice

logger = logging.getLogger(__name__)


class APIProvider(TypedDict, total=False):
    name: Required[str]
    base_url: Required[str | None]
    api_key: Required[str | None]
    # Wildcard patterns for model names that support response schema validation:
    response_schema_support: tuple[str, ...] | None


class CloudLLMSettings(LLMSettings, total=False):
    extra_headers: dict[str, Any] | None
    extra_body: object | None
    extra_query: dict[str, Any] | None


LLMRateLimiter = RateLimiter[
    AssistantMessage
    | AsyncIterator[
        CompletionChunkEvent[CompletionChunk] | CompletionEvent | LLMStreamingErrorEvent
    ],
]


@dataclass(frozen=True)
class CloudLLM(LLM):
    llm_settings: CloudLLMSettings | None = None
    api_provider: APIProvider | None = None
    rate_limiter: LLMRateLimiter | None = None
    apply_response_schema_via_provider: bool = False
    apply_tool_call_schema_via_provider: bool = False
    http_client: httpx.AsyncClient | None = None

    def __post_init__(self) -> None:
        if self.rate_limiter is not None:
            logger.info(
                f"[{self.__class__.__name__}] Set rate limit to "
                f"{self.rate_limiter.rpm} RPM"
            )

        if self.apply_response_schema_via_provider:
            object.__setattr__(self, "apply_tool_call_schema_via_provider", True)

    @abstractmethod
    async def _get_api_completion(
        self,
        api_messages: list[Any],
        *,
        api_tools: list[Any] | None = None,
        api_tool_choice: Any | None = None,
        api_response_schema: type | None = None,
        **api_llm_settings: Any,
    ) -> Any:
        pass

    @abstractmethod
    async def _get_api_completion_stream(
        self,
        api_messages: list[Any],
        *,
        api_tools: list[Any] | None = None,
        api_tool_choice: Any | None = None,
        api_response_schema: type | None = None,
        **api_llm_settings: Any,
    ) -> AsyncIterator[Any]:
        pass

    def __init_subclass__(cls, **kwargs: Any):
        super().__init_subclass__(**kwargs)

        if "_get_api_completion" in cls.__dict__:
            cls._get_api_completion = limit_rate(cls._get_api_completion)  # type: ignore[method-assign]

        if "_get_api_completion_stream" in cls.__dict__:
            cls._get_api_completion_stream = limit_rate(cls._get_api_completion_stream)  # type: ignore[method-assign]

    def _make_api_completion_kwargs(
        self,
        messages: Messages,
        response_schema: Any | None = None,
        tools: Mapping[str, BaseTool[BaseModel, Any, Any]] | None = None,
        tool_choice: ToolChoice | None = None,
        **extra_llm_settings: Any,
    ) -> dict[str, Any]:
        api_messages = [self.converters.to_message(m) for m in messages]

        api_tools = None
        api_tool_choice = None
        if tools:
            strict = True if self.apply_tool_call_schema_via_provider else None
            api_tools = [
                self.converters.to_tool(t, strict=strict) for t in tools.values()
            ]
            if tool_choice is not None:
                api_tool_choice = self.converters.to_tool_choice(tool_choice)

        api_llm_settings = deepcopy((self.llm_settings or {}) | extra_llm_settings)

        return dict(
            api_messages=api_messages,
            api_tools=api_tools,
            api_tool_choice=api_tool_choice,
            api_response_schema=response_schema,
            **api_llm_settings,
        )

    async def _generate_completion_once(
        self,
        messages: Messages,
        *,
        response_schema: Any | None = None,
        response_schema_by_xml_tag: Mapping[str, Any] | None = None,
        tools: Mapping[str, BaseTool[BaseModel, Any, Any]] | None = None,
        tool_choice: ToolChoice | None = None,
        **extra_llm_settings: Any,
    ) -> Completion:
        completion_kwargs = self._make_api_completion_kwargs(
            messages=messages,
            response_schema=response_schema,
            tools=tools,
            tool_choice=tool_choice,
            **extra_llm_settings,
        )

        if not self.apply_response_schema_via_provider:
            completion_kwargs.pop("api_response_schema", None)
        api_completion = await self._get_api_completion(**completion_kwargs)

        completion = self.converters.from_completion(api_completion, name=self.model_id)

        # if not self.apply_response_schema_via_provider:
        self._validate_response(
            completion,
            response_schema=response_schema,
            response_schema_by_xml_tag=response_schema_by_xml_tag,
        )
        # if not self.apply_tool_call_schema_via_provider and tools is not None:
        if tools is not None:
            self._validate_tool_calls(completion, tools=tools)

        return completion

    async def _generate_completion_stream_once(
        self,
        messages: Messages,
        *,
        response_schema: Any | None = None,
        response_schema_by_xml_tag: Mapping[str, Any] | None = None,
        tools: Mapping[str, BaseTool[BaseModel, Any, Any]] | None = None,
        tool_choice: ToolChoice | None = None,
        proc_name: str | None = None,
        call_id: str | None = None,
        **extra_llm_settings: Any,
    ) -> AsyncIterator[CompletionChunkEvent[CompletionChunk] | CompletionEvent]:
        completion_kwargs = self._make_api_completion_kwargs(
            messages=messages,
            response_schema=response_schema,
            tools=tools,
            tool_choice=tool_choice,
            **extra_llm_settings,
        )
        if not self.apply_response_schema_via_provider:
            completion_kwargs.pop("api_response_schema", None)

        api_stream = await self._get_api_completion_stream(**completion_kwargs)

        api_completion_chunks: list[Any] = []

        async for api_completion_chunk in api_stream:
            api_completion_chunks.append(api_completion_chunk)
            completion_chunk = self.converters.from_completion_chunk(
                api_completion_chunk, name=self.model_id
            )

            yield CompletionChunkEvent(
                data=completion_chunk, proc_name=proc_name, call_id=call_id
            )

        api_completion = self.combine_completion_chunks(
            api_completion_chunks, response_schema=response_schema, tools=tools
        )
        completion = self.converters.from_completion(api_completion, name=self.model_id)

        yield CompletionEvent(data=completion, proc_name=proc_name, call_id=call_id)

        # if not self.apply_response_schema_via_provider:
        self._validate_response(
            completion,
            response_schema=response_schema,
            response_schema_by_xml_tag=response_schema_by_xml_tag,
        )
        # if not self.apply_tool_call_schema_via_provider and tools is not None:
        if tools is not None:
            self._validate_tool_calls(completion, tools=tools)



from __future__ import annotations

from typing import Any, ClassVar, Self, TypeVar, cast, get_args


class AutoInstanceAttributesMixin:
    """
    A **runtime convenience mix-in** that automatically exposes the concrete
    types supplied to a *generic* base class as **instance attributes**.

    Example:
    -------
    from typing import Generic, TypeVar
    from grasp_agents.generics_utils import AutoInstanceAttributesMixin

    T = TypeVar("T")
    U = TypeVar("U")

    class MyBase(AutoInstanceAttributesMixin, Generic[T, U]):
        _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
            0: "elem_type",
            1: "meta_type",
        }

    class Packet(MyBase[int, str]):
        ...

    Alias = MyBase[bytes, float]   # "late" specialization

    print(Packet().elem_type)      # <class 'int'>
    print(Alias().meta_type)       # <class 'float'>

    """

    # Configure this on your *generic* base class
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {}

    # Filled automatically for every concrete specialization
    _resolved_instance_attr_types: ClassVar[dict[str, type]]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._set_resolved_generic_instance_attributes()

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Runs when a *class statement* defines a new subclass."""
        super().__init_subclass__(**kwargs)
        cls._resolved_instance_attr_types = cls._compute_resolved_attrs(cls)

    @classmethod
    def __class_getitem__(cls, params: Any) -> type[Self]:
        """
        Run when someone writes ``SomeGeneric[ConcreteTypes]``.
        Ensures aliases receive the same resolved-type mapping.
        """
        specialized: type[Self] = cast("type[Self]", super().__class_getitem__(params))  # type: ignore[assignment]
        specialized._resolved_instance_attr_types = cls._compute_resolved_attrs(  # noqa: SLF001
            specialized
        )

        return specialized

    @staticmethod
    def _compute_resolved_attrs(_cls: type) -> dict[str, type]:
        """
        Walks the MRO, finds the first generic base that defines
        `_generic_arg_to_instance_attr_map`, and resolves concrete types.
        """
        target_generic_base: Any | None = None
        attr_mapping: dict[int, str] | None = None

        # Locate the mapping
        for mro_cls in _cls.mro():
            if mro_cls in {AutoInstanceAttributesMixin, object}:
                continue
            if (
                hasattr(mro_cls, "__parameters__")
                and mro_cls.__parameters__  # type: ignore[has-type]
                and "_generic_arg_to_instance_attr_map" in mro_cls.__dict__
            ):
                target_generic_base = mro_cls
                attr_mapping = cast(
                    "dict[int, str]",
                    mro_cls._generic_arg_to_instance_attr_map,  # type: ignore[attr-defined] # noqa: SLF001
                )
                break

        if target_generic_base is None or attr_mapping is None:
            return {}

        resolved: dict[str, type] = {}

        def _add_to_resolved(generic_args: tuple[type | TypeVar, ...]) -> None:
            for index, attr_name in attr_mapping.items():
                if attr_name in resolved:
                    continue
                if index < len(generic_args):
                    arg = generic_args[index]
                    if not isinstance(arg, TypeVar):
                        resolved[attr_name] = arg

        def _all_resolved() -> bool:
            return all(name in resolved for name in attr_mapping.values())

        # Scenario 1: _cls itself is the direct parameterization (handles aliases).
        # e.g., _cls is MyBase[bytes, float]. Its __origin__ is MyBase.
        if getattr(_cls, "__origin__", None) is target_generic_base:
            _add_to_resolved(get_args(_cls))

        # Scenario 2: Check MRO for subclasses or more complex structures.
        # This also acts as a fallback if Scenario 1 didn't resolve all attributes.
        if not _all_resolved():
            for mro_candidate in _cls.mro():
                # Pydantic-specific check first
                pydantic_generic_metadata = getattr(
                    mro_candidate, "__pydantic_generic_metadata__", None
                )
                if (
                    pydantic_generic_metadata
                    and pydantic_generic_metadata.get("origin") is target_generic_base
                ):
                    _add_to_resolved(pydantic_generic_metadata.get("args", ()))
                    if _all_resolved():
                        break

                # Fallback to standard generic introspection if not fully resolved by Pydantic check
                if not _all_resolved():
                    mro_candidate_origin = getattr(mro_candidate, "__origin__", None)
                    if mro_candidate_origin is target_generic_base:
                        _add_to_resolved(get_args(mro_candidate))
                        if _all_resolved():
                            break

                if not _all_resolved():
                    mro_candidate_orig_bases = getattr(
                        mro_candidate, "__orig_bases__", []
                    )
                    for param_base in mro_candidate_orig_bases:
                        param_base_origin = getattr(param_base, "__origin__", None)
                        if param_base_origin is target_generic_base:
                            _add_to_resolved(get_args(param_base))
                            if _all_resolved():
                                break

                if _all_resolved():
                    break

        return resolved

    def _set_resolved_generic_instance_attributes(self) -> None:
        attr_names = self._generic_arg_to_instance_attr_map.values()
        resolved_attr_types = getattr(
            self.__class__, "_resolved_instance_attr_types", {}
        )
        pyd_private = getattr(self, "__pydantic_private__", {})

        for attr_name in attr_names:
            if attr_name in resolved_attr_types:
                attr_type = resolved_attr_types[attr_name]
                # attr_type = None if _attr_type is type(None) else _attr_type
            else:
                attr_type = object

            if attr_name in pyd_private:
                pyd_private[attr_name] = attr_type
            else:
                setattr(self, attr_name, attr_type)



import ast
import json
import re
from collections.abc import Mapping
from logging import getLogger
from typing import Annotated, Any, NoReturn, TypeVar, get_args, get_origin

from pydantic import TypeAdapter
from pydantic import ValidationError as PydanticValidationError

from .errors import JSONSchemaValidationError, PyJSONStringParsingError

logger = getLogger(__name__)

_JSON_START_RE = re.compile(r"[{\[]")


def extract_json_substring(text: str) -> str | None:
    decoder = json.JSONDecoder()
    for match in _JSON_START_RE.finditer(text):
        start = match.start()
        try:
            _, end = decoder.raw_decode(text, idx=start)
            return text[start:end]
        except json.JSONDecodeError:
            continue

    return None


def parse_json_or_py_string(
    s: str,
    from_substring: bool = False,
    return_none_on_failure: bool = False,
    strip_language_markdown: bool = True,
) -> dict[str, Any] | list[Any] | None:
    s_orig = s

    if strip_language_markdown:
        s = re.sub(r"```[a-zA-Z0-9]*\n|```", "", s).strip()

    if from_substring:
        s = extract_json_substring(s) or ""

    try:
        return ast.literal_eval(s)
    except (ValueError, SyntaxError):
        try:
            return json.loads(s)
        except json.JSONDecodeError as exc:
            err_message = (
                "Both ast.literal_eval and json.loads "
                f"failed to parse the following JSON/Python string:\n{s_orig}"
            )
            if return_none_on_failure:
                logger.warning(err_message)
                return None
            raise PyJSONStringParsingError(s_orig, message=err_message) from exc


def is_str_type(t: Any) -> bool:
    type_origin = get_origin(t)
    type_args = get_args(t)

    return (t is str) or (
        (type_origin is Annotated) and len(type_args) > 0 and type_args[0] is str
    )


T = TypeVar("T")


def validate_obj_from_json_or_py_string(
    s: str,
    schema: type[T],
    from_substring: bool = False,
    strip_language_markdown: bool = True,
) -> T:
    try:
        if is_str_type(schema):
            parsed = s
        else:
            parsed = parse_json_or_py_string(
                s,
                return_none_on_failure=True,
                from_substring=from_substring,
                strip_language_markdown=strip_language_markdown,
            )
        return TypeAdapter(schema).validate_python(parsed)
    except PydanticValidationError as exc:
        raise JSONSchemaValidationError(s, schema) from exc


def validate_tagged_objs_from_json_or_py_string(
    s: str,
    schema_by_xml_tag: Mapping[str, type[T]],
    from_substring: bool = False,
    strip_language_markdown: bool = True,
) -> Mapping[str, T]:
    validated_obj_per_tag: dict[str, T] = {}
    _schema: type[T] | None = None
    _tag: str | None = None

    def _raise(
        tag: str,
        schema: Any,
        *,
        not_found: bool = False,
        base_exc: Exception | None = None,
    ) -> NoReturn:
        if not_found:
            msg = (
                f"Failed to find valid tagged section <{tag}>...</{tag}> "
                f"in the string:\n{s}"
            )
        else:
            msg = (
                f"Failed to validate substring within tag <{tag}> against JSON schema:"
                f"\n{s}\nExpected type: {schema}"
            )
        raise JSONSchemaValidationError(s, schema, message=msg) from base_exc

    for _tag, _schema in schema_by_xml_tag.items():
        if f"<{_tag}>" in s:
            match = re.search(rf"<{_tag}>\s*(.*?)\s*</{_tag}>", s, re.DOTALL)
            if match is None:
                _raise(_tag, _schema, not_found=True, base_exc=None)
            try:
                tagged_substring = match.group(1).strip()
                validated_obj_per_tag[_tag] = validate_obj_from_json_or_py_string(
                    tagged_substring,  # type: ignore[assignment]
                    schema=_schema,
                    from_substring=from_substring,
                    strip_language_markdown=strip_language_markdown,
                )
            except JSONSchemaValidationError as exc:
                _raise(_tag, _schema, base_exc=exc)

    return validated_obj_per_tag



import asyncio
import logging
from collections.abc import AsyncIterator, Sequence
from typing import Any, ClassVar, Generic, cast

from grasp_agents.tracing_decorators import agent

from ..memory import MemT
from ..packet import Packet
from ..run_context import CtxT, RunContext
from ..typing.events import Event, ProcPacketOutputEvent, ProcPayloadOutputEvent
from ..typing.io import InT, OutT
from ..utils import stream_concurrent
from .base_processor import BaseProcessor, with_retry, with_retry_stream

logger = logging.getLogger(__name__)


class ParallelProcessor(
    BaseProcessor[InT, OutT, MemT, CtxT], Generic[InT, OutT, MemT, CtxT]
):
    """Processor that runs multiple inputs in parallel, each producing one output."""

    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    async def _process(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: InT | None = None,
        memory: MemT,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> OutT:
        """Process a single input and return a single output."""
        return cast("OutT", in_args)

    async def _process_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: InT | None = None,
        memory: MemT,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> AsyncIterator[Event[Any]]:
        output = cast("OutT", in_args)
        yield ProcPayloadOutputEvent(data=output, proc_name=self.name, call_id=call_id)

    @with_retry
    async def _run_single(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: InT | None = None,
        forgetful: bool = False,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> Packet[OutT]:
        memory = self.memory.model_copy(deep=True) if forgetful else self.memory

        output = await self._process(
            chat_inputs=chat_inputs,
            in_args=in_args,
            memory=memory,
            call_id=call_id,
            ctx=ctx,
        )
        val_output = self._validate_output(output, call_id=call_id)

        recipients = self.select_recipients(output=val_output, ctx=ctx, call_id=call_id)

        return Packet(
            payloads=[val_output],
            sender=self.name,
            routing=[recipients] if recipients is not None else None,
        )

    def _join_payloads(self, packets: Sequence[Packet[OutT]]) -> list[OutT]:
        return [p.payloads[0] for p in packets]

    def _join_routings(
        self, packets: Sequence[Packet[OutT]]
    ) -> Sequence[Sequence[str]] | None:
        if not packets:
            return None
        routings = [p.routing[0] if p.routing is not None else None for p in packets]
        if all(r is None for r in routings):
            joined_routing = None
        else:
            joined_routing = [r or [] for r in routings]

        return joined_routing

    async def _run_parallel(
        self, in_args: list[InT], call_id: str, ctx: RunContext[CtxT]
    ) -> Packet[OutT]:
        tasks = [
            self._run_single(
                in_args=inp, forgetful=True, call_id=f"{call_id}/{idx}", ctx=ctx
            )
            for idx, inp in enumerate(in_args)
        ]
        out_packets = await asyncio.gather(*tasks)

        return Packet(
            sender=self.name,
            payloads=self._join_payloads(out_packets),
            routing=self._join_routings(out_packets),
        )

    @agent(name="processor")  # type: ignore
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
        forgetful: bool = False,
        call_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> Packet[OutT]:
        call_id = self._generate_call_id(call_id)
        ctx = ctx or RunContext[CtxT](state=None)  # type: ignore

        val_in_args = self._validate_inputs(
            call_id=call_id,
            chat_inputs=chat_inputs,
            in_packet=in_packet,
            in_args=in_args,
        )

        if val_in_args and len(val_in_args) > 1:
            return await self._run_parallel(
                in_args=val_in_args, call_id=call_id, ctx=ctx
            )

        return await self._run_single(
            chat_inputs=chat_inputs,
            in_args=val_in_args[0] if val_in_args else None,
            forgetful=forgetful,
            call_id=call_id,
            ctx=ctx,
        )

    @with_retry_stream
    async def _run_single_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: InT | None = None,
        forgetful: bool = False,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> AsyncIterator[Event[Any]]:
        memory = self.memory.model_copy(deep=True) if forgetful else self.memory

        output: OutT | None = None
        async for event in self._process_stream(
            chat_inputs=chat_inputs,
            in_args=in_args,
            memory=memory,
            call_id=call_id,
            ctx=ctx,
        ):
            if isinstance(event, ProcPayloadOutputEvent):
                output = event.data
            yield event

        assert output is not None

        val_output = self._validate_output(output, call_id=call_id)

        recipients = self.select_recipients(output=val_output, ctx=ctx, call_id=call_id)

        out_packet = Packet[OutT](
            payloads=[val_output],
            sender=self.name,
            routing=[recipients] if recipients is not None else None,
        )

        yield ProcPacketOutputEvent(
            data=out_packet, proc_name=self.name, call_id=call_id
        )

    async def _run_parallel_stream(
        self,
        in_args: list[InT],
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> AsyncIterator[Event[Any]]:
        streams = [
            self._run_single_stream(
                in_args=inp, forgetful=True, call_id=f"{call_id}/{idx}", ctx=ctx
            )
            for idx, inp in enumerate(in_args)
        ]

        out_packets_map: dict[int, Packet[OutT]] = {}
        async for idx, event in stream_concurrent(streams):
            if isinstance(event, ProcPacketOutputEvent):
                out_packets_map[idx] = event.data
            else:
                yield event

        out_packet = Packet(
            sender=self.name,
            payloads=self._join_payloads(list(out_packets_map.values())),
            routing=self._join_routings(list(out_packets_map.values())),
        )

        yield ProcPacketOutputEvent(
            data=out_packet, proc_name=self.name, call_id=call_id
        )

    @agent(name="processor")  # type: ignore
    async def run_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
        forgetful: bool = False,
        call_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> AsyncIterator[Event[Any]]:
        call_id = self._generate_call_id(call_id)
        ctx = ctx or RunContext[CtxT](state=None)  # type: ignore

        val_in_args = self._validate_inputs(
            call_id=call_id,
            chat_inputs=chat_inputs,
            in_packet=in_packet,
            in_args=in_args,
        )

        if val_in_args and len(val_in_args) > 1:
            stream = self._run_parallel_stream(
                in_args=val_in_args, call_id=call_id, ctx=ctx
            )
        else:
            stream = self._run_single_stream(
                chat_inputs=chat_inputs,
                in_args=val_in_args[0] if val_in_args else None,
                forgetful=forgetful,
                call_id=call_id,
                ctx=ctx,
            )
        async for event in stream:
            yield event



from collections.abc import Sequence

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult

# Set of LLM provider names used in OpenTelemetry attributes
# See https://opentelemetry.io/docs/specs/semconv/registry/attributes/gen-ai/#gen-ai-provider-name
LLM_PROVIDER_NAMES = {
    "anthropic",
    "aws.bedrock",
    "azure.ai.inference",
    "azure.ai.openai",
    "cohere",
    "deepseek",
    "gcp.gemini",
    "gcp.gen_ai",
    "gcp.vertex_ai",
    "groq",
    "ibm.watsonx.ai",
    "mistral_ai",
    "openai",
    "perplexity",
    "x_ai",
}


class NoopExporter(SpanExporter):
    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        return SpanExportResult.SUCCESS


class FilteringExporter(SpanExporter):
    def __init__(self, inner: SpanExporter, blocklist: set[str] | None = None):
        self._inner = inner
        self._blocklist = blocklist or set()

    def export(self, spans: Sequence[ReadableSpan]):
        keep: list[ReadableSpan] = []
        for s in spans:
            attrs = s.attributes or {}
            provider = attrs.get("gen_ai.system", "") or attrs.get(
                "gen_ai.provider.name", ""
            )
            if provider not in self._blocklist:
                keep.append(s)

        return SpanExportResult.SUCCESS if not keep else self._inner.export(keep)

    def shutdown(self):
        self._inner.shutdown()

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        return self._inner.force_flush(timeout_millis)


class LogScopeExporter(SpanExporter):
    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        for s in spans:
            scope = getattr(s, "instrumentation_scope", None)
            print(
                "SCOPE:",
                getattr(scope, "name", ""),
                "SPAN:",
                s.name,
                "ATTRS:",
                s.attributes,
            )
        return SpanExportResult.SUCCESS



from collections.abc import Sequence
from typing import Any, Generic, Self, TypeVar
from uuid import uuid4

from pydantic import BaseModel, ConfigDict, Field, model_validator

from .typing.io import ProcName

PacketRouting = Sequence[Sequence[ProcName]]
_PayloadT_co = TypeVar("_PayloadT_co", covariant=True)


def is_uniform_routing(routing: PacketRouting | None) -> Sequence[ProcName] | None:
    if not routing or len(routing) == 0:
        return None

    first_recipients = routing[0]
    first_set = set(first_recipients)
    for recipients in routing[1:]:
        if len(recipients) != len(first_recipients) or set(recipients) != first_set:
            return None

    return first_recipients


class Packet(BaseModel, Generic[_PayloadT_co]):
    id: str = Field(default_factory=lambda: str(uuid4())[:8])
    payloads: Sequence[_PayloadT_co]
    sender: ProcName
    routing: PacketRouting | None = None

    @property
    def uniform_routing(self) -> Sequence[ProcName] | None:
        return is_uniform_routing(self.routing)

    @model_validator(mode="before")
    @classmethod
    def _normalize_routing(cls, data: dict[str, Any]) -> dict[str, Any]:
        routing = data.get("routing")
        if (
            routing
            and isinstance(routing, (list, tuple))
            and all(isinstance(r, str) for r in routing)  # type: ignore[misc]
        ):
            payloads = data.get("payloads", [])
            data["routing"] = [routing for _ in range(len(payloads))]
        return data

    @model_validator(mode="after")
    def _validate_routing(self) -> Self:
        if self.routing is not None and len(self.payloads) != len(self.routing):
            raise ValueError(
                "If routing is specified, its length must match the length of payloads"
            )
        return self

    def split_per_payload(self) -> Sequence["Packet[_PayloadT_co]"] | None:
        if self.routing is None:
            return None

        single_payload = len(self.payloads) == 1

        return [
            Packet(
                id=f"{self.id}/{i}" if not single_payload else self.id,
                payloads=[payload],
                routing=[recipients],
                sender=self.sender,
            )
            for i, (payload, recipients) in enumerate(
                zip(self.payloads, self.routing, strict=False)
            )
        ]

    def split_by_recipient(
        self,
    ) -> Sequence["Packet[_PayloadT_co]"] | None:
        if self.routing is None:
            return None

        recipient_to_payloads: dict[ProcName, list[_PayloadT_co]] = {}
        for payload, recipients in zip(self.payloads, self.routing, strict=True):
            for recipient in recipients:
                if recipient not in recipient_to_payloads:
                    recipient_to_payloads[recipient] = []
                recipient_to_payloads[recipient].append(payload)

        single_recipient = len(recipient_to_payloads) == 1
        return [
            Packet(
                id=f"{self.id}/{recipient}" if not single_recipient else self.id,
                payloads=payloads,
                routing=[[recipient] for _ in range(len(payloads))],
                sender=self.sender,
            )
            for recipient, payloads in recipient_to_payloads.items()
        ]

    model_config = ConfigDict(extra="forbid")

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}:\n"
            f"ID: {self.id}\n"
            f"From: {self.sender}\n"
            f"Routing: {self.routing or 'None'}\n"
            f"Payloads: {len(self.payloads)}"
        )



import json
from collections.abc import Hashable, Mapping, Sequence
from enum import StrEnum
from typing import Annotated, Any, Literal, Required, TypeAlias
from uuid import uuid4

from litellm.types.llms.openai import ChatCompletionAnnotation as LiteLLMAnnotation
from pydantic import BaseModel, Field
from pydantic.json import pydantic_encoder
from typing_extensions import TypedDict

from .content import Content, ImageData
from .tool import ToolCall


class Role(StrEnum):
    USER = "user"
    SYSTEM = "system"
    DEVELOPER = "developer"
    ASSISTANT = "assistant"
    TOOL = "tool"


class MessageBase(BaseModel):
    id: Hashable = Field(default_factory=lambda: str(uuid4())[:8])
    name: str | None = None


class ChatCompletionCachedContent(TypedDict):
    type: Literal["ephemeral"]


class ThinkingBlock(TypedDict, total=False):
    type: Required[Literal["thinking"]]
    thinking: str
    signature: str | None
    cache_control: dict[str, Any] | ChatCompletionCachedContent | None


class RedactedThinkingBlock(TypedDict, total=False):
    type: Required[Literal["redacted_thinking"]]
    data: str
    cache_control: dict[str, Any] | ChatCompletionCachedContent | None


class AssistantMessage(MessageBase):
    role: Literal[Role.ASSISTANT] = Role.ASSISTANT
    content: str | None
    tool_calls: Sequence[ToolCall] | None = None
    refusal: str | None = None
    reasoning_content: str | None = None
    thinking_blocks: Sequence[ThinkingBlock | RedactedThinkingBlock] | None = None
    annotations: Sequence[LiteLLMAnnotation] | None = None
    provider_specific_fields: dict[str, Any] | None = None


class UserMessage(MessageBase):
    role: Literal[Role.USER] = Role.USER
    content: Content | str

    @classmethod
    def from_text(cls, text: str, name: str | None = None) -> "UserMessage":
        return cls(content=Content.from_text(text), name=name)

    @classmethod
    def from_image(cls, image: ImageData, name: str | None = None) -> "UserMessage":
        return cls(content=Content.from_image(image), name=name)

    @classmethod
    def from_images(
        cls, images: Sequence[ImageData], name: str | None = None
    ) -> "UserMessage":
        return cls(content=Content.from_images(images), name=name)

    @classmethod
    def from_content_parts(
        cls,
        content_parts: Sequence[str | ImageData],
        name: str | None = None,
    ) -> "UserMessage":
        return cls(content=Content.from_content_parts(content_parts), name=name)

    @classmethod
    def from_formatted_prompt(
        cls,
        prompt_template: str,
        name: str | None = None,
        prompt_args: Mapping[str, str | int | bool | ImageData] | None = None,
    ) -> "UserMessage":
        content = Content.from_formatted_prompt(prompt_template, **(prompt_args or {}))

        return cls(content=content, name=name)


class SystemMessage(MessageBase):
    role: Literal[Role.SYSTEM] = Role.SYSTEM
    content: str


class ToolMessage(MessageBase):
    role: Literal[Role.TOOL] = Role.TOOL
    content: str
    tool_call_id: str

    @classmethod
    def from_tool_output(
        cls, tool_output: Any, tool_call: ToolCall, indent: int = 2
    ) -> "ToolMessage":
        return cls(
            content=json.dumps(tool_output, default=pydantic_encoder, indent=indent),
            tool_call_id=tool_call.id,
            name=tool_call.tool_name,
        )


Message = Annotated[
    AssistantMessage | UserMessage | SystemMessage | ToolMessage,
    Field(discriminator="role"),
]

Messages: TypeAlias = list[Message]



from collections.abc import AsyncIterator, Mapping, Sequence
from pathlib import Path
from typing import Any, ClassVar, Generic, Protocol, TypeVar, cast, final

from pydantic import BaseModel

from .llm import LLM
from .llm_agent_memory import LLMAgentMemory
from .llm_policy_executor import (
    AfterGenerateHook,
    BeforeGenerateHook,
    FinalAnswerChecker,
    LLMPolicyExecutor,
    ToolOutputConverter,
)
from .processors.parallel_processor import ParallelProcessor
from .prompt_builder import (
    InputContentBuilder,
    PromptBuilder,
    SystemPromptBuilder,
)
from .run_context import CtxT, RunContext
from .typing.content import Content, ImageData
from .typing.events import (
    Event,
    ProcPayloadOutputEvent,
    SystemMessageEvent,
    UserMessageEvent,
)
from .typing.io import InT, LLMPrompt, OutT, ProcName
from .typing.message import Message, SystemMessage, UserMessage
from .typing.tool import BaseTool, ToolCall
from .utils import get_prompt, is_method_overridden
from .validation import validate_obj_from_json_or_py_string

_InT_contra = TypeVar("_InT_contra", contravariant=True)
_OutT_co = TypeVar("_OutT_co", covariant=True)


class MemoryPreparator(Protocol):
    def __call__(
        self,
        memory: "LLMAgentMemory",
        *,
        instructions: LLMPrompt | None = None,
        in_args: Any | None,
        ctx: RunContext[Any],
        call_id: str,
    ) -> None: ...


class OutputParser(Protocol[_InT_contra, _OutT_co, CtxT]):
    def __call__(
        self,
        final_answer: str,
        *,
        memory: LLMAgentMemory,
        in_args: _InT_contra | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> _OutT_co: ...


class LLMAgent(
    ParallelProcessor[InT, OutT, LLMAgentMemory, CtxT],
    Generic[InT, OutT, CtxT],
):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self,
        name: ProcName,
        *,
        # LLM
        llm: LLM,
        # Tools
        tools: list[BaseTool[Any, Any, CtxT]] | None = None,
        # Input prompt template (combines user and received arguments)
        in_prompt: LLMPrompt | None = None,
        in_prompt_path: str | Path | None = None,
        # System prompt template
        sys_prompt: LLMPrompt | None = None,
        sys_prompt_path: str | Path | None = None,
        # LLM response validation
        response_schema: Any | None = None,
        response_schema_by_xml_tag: Mapping[str, Any] | None = None,
        # Agent loop settings
        max_turns: int = 100,
        react_mode: bool = False,
        final_answer_as_tool_call: bool = False,
        # Agent memory management
        memory: LLMAgentMemory | None = None,
        reset_memory_on_run: bool = False,
        # Agent run retries
        max_retries: int = 0,
        # Multi-agent routing
        recipients: Sequence[ProcName] | None = None,
    ) -> None:
        super().__init__(name=name, recipients=recipients, max_retries=max_retries)

        # Agent memory

        self._memory: LLMAgentMemory = memory or LLMAgentMemory()
        self._reset_memory_on_run = reset_memory_on_run

        # LLM policy executor

        if issubclass(self._out_type, BaseModel):
            final_answer_type = self._out_type
        elif not final_answer_as_tool_call:
            final_answer_type = BaseModel
        else:
            raise TypeError(
                "Final answer type must be a subclass of BaseModel if "
                "final_answer_as_tool_call is True."
            )

        self._used_default_llm_response_schema: bool = False
        if (
            response_schema is None
            and tools is None
            and not is_method_overridden(
                "parse_output_impl", self, LLMAgent[Any, Any, Any]
            )
        ):
            response_schema = self.out_type
            self._used_default_llm_response_schema = True

        self._policy_executor: LLMPolicyExecutor[CtxT] = LLMPolicyExecutor[CtxT](
            agent_name=self.name,
            llm=llm,
            tools=tools,
            response_schema=response_schema,
            response_schema_by_xml_tag=response_schema_by_xml_tag,
            max_turns=max_turns,
            react_mode=react_mode,
            final_answer_type=final_answer_type,
            final_answer_as_tool_call=final_answer_as_tool_call,
        )

        # Prompt builder

        sys_prompt = get_prompt(prompt_text=sys_prompt, prompt_path=sys_prompt_path)
        in_prompt = get_prompt(prompt_text=in_prompt, prompt_path=in_prompt_path)

        self._prompt_builder = PromptBuilder[self.in_type, CtxT](
            agent_name=self._name, sys_prompt=sys_prompt, in_prompt=in_prompt
        )

        self._register_overridden_implementations()

    @property
    def llm(self) -> LLM:
        return self._policy_executor.llm

    @property
    def tools(self) -> dict[str, BaseTool[BaseModel, Any, CtxT]]:
        return self._policy_executor.tools

    @property
    def max_turns(self) -> int:
        return self._policy_executor.max_turns

    @property
    def sys_prompt(self) -> LLMPrompt | None:
        return self._prompt_builder.sys_prompt

    @property
    def in_prompt(self) -> LLMPrompt | None:
        return self._prompt_builder.in_prompt

    @final
    def prepare_memory(
        self,
        memory: LLMAgentMemory,
        *,
        instructions: LLMPrompt | None = None,
        in_args: InT | None = None,
        ctx: RunContext[Any],
        call_id: str,
    ) -> None:
        if is_method_overridden("prepare_memory_impl", self, LLMAgent[Any, Any, Any]):
            return self.prepare_memory_impl(
                memory=memory,
                instructions=instructions,
                in_args=in_args,
                ctx=ctx,
                call_id=call_id,
            )

    def _memorize_inputs(
        self,
        memory: LLMAgentMemory,
        *,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        in_args: InT | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> tuple[SystemMessage | None, UserMessage | None]:
        formatted_sys_prompt = self._prompt_builder.build_system_prompt(
            ctx=ctx, call_id=call_id
        )

        system_message: SystemMessage | None = None
        if self._reset_memory_on_run or memory.is_empty:
            memory.reset(formatted_sys_prompt)
            if formatted_sys_prompt is not None:
                system_message = cast("SystemMessage", memory.messages[0])
        else:
            self.prepare_memory(
                memory=memory,
                instructions=formatted_sys_prompt,
                in_args=in_args,
                ctx=ctx,
                call_id=call_id,
            )

        input_message = self._prompt_builder.build_input_message(
            chat_inputs=chat_inputs, in_args=in_args, ctx=ctx, call_id=call_id
        )
        if input_message:
            memory.update([input_message])

        return system_message, input_message

    def parse_output_default(self, final_answer: str) -> OutT:
        return validate_obj_from_json_or_py_string(
            final_answer,
            schema=self._out_type,
            from_substring=False,
            strip_language_markdown=True,
        )

    @final
    def parse_output(
        self,
        final_answer: str,
        *,
        memory: LLMAgentMemory,
        in_args: InT | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> OutT:
        if is_method_overridden("parse_output_impl", self, LLMAgent[Any, Any, Any]):
            return self.parse_output_impl(
                final_answer,
                memory=memory,
                in_args=in_args,
                ctx=ctx,
                call_id=call_id,
            )

        return self.parse_output_default(final_answer)

    async def _process(
        self,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        *,
        in_args: InT | None = None,
        memory: LLMAgentMemory,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> OutT:
        system_message, input_message = self._memorize_inputs(
            memory=memory,
            chat_inputs=chat_inputs,
            in_args=in_args,
            ctx=ctx,
            call_id=call_id,
        )
        if system_message:
            self._print_messages([system_message], ctx=ctx, call_id=call_id)
        if input_message:
            self._print_messages([input_message], ctx=ctx, call_id=call_id)

        final_answer = await self._policy_executor.execute(
            memory, ctx=ctx, call_id=call_id
        )

        return self.parse_output(
            final_answer,
            memory=memory,
            in_args=in_args,
            ctx=ctx,
            call_id=call_id,
        )

    async def _process_stream(
        self,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        *,
        in_args: InT | None = None,
        memory: LLMAgentMemory,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> AsyncIterator[Event[Any]]:
        system_message, input_message = self._memorize_inputs(
            memory=memory,
            chat_inputs=chat_inputs,
            in_args=in_args,
            ctx=ctx,
            call_id=call_id,
        )
        if system_message:
            self._print_messages([system_message], ctx=ctx, call_id=call_id)
            yield SystemMessageEvent(
                data=system_message, proc_name=self.name, call_id=call_id
            )
        if input_message:
            self._print_messages([input_message], ctx=ctx, call_id=call_id)
            yield UserMessageEvent(
                data=input_message, proc_name=self.name, call_id=call_id
            )

        event: Event[Any] | None = None
        async for event in self._policy_executor.execute_stream(
            memory, ctx=ctx, call_id=call_id
        ):
            yield event

        final_answer = self._policy_executor.get_final_answer(memory)

        output = self.parse_output(
            final_answer or "",
            memory=memory,
            in_args=in_args,
            ctx=ctx,
            call_id=call_id,
        )
        yield ProcPayloadOutputEvent(data=output, proc_name=self.name, call_id=call_id)

    def _print_messages(
        self, messages: Sequence[Message], ctx: RunContext[CtxT], call_id: str
    ) -> None:
        if ctx.printer:
            ctx.printer.print_messages(messages, agent_name=self.name, call_id=call_id)

    # Methods that can be overridden in subclasses

    def prepare_memory_impl(
        self,
        memory: LLMAgentMemory,
        *,
        instructions: LLMPrompt | None = None,
        in_args: InT | None = None,
        ctx: RunContext[Any],
        call_id: str,
    ) -> None:
        raise NotImplementedError

    def parse_output_impl(
        self,
        final_answer: str,
        *,
        memory: LLMAgentMemory,
        in_args: InT | None = None,
        ctx: RunContext[CtxT],
        call_id: str,
    ) -> OutT:
        raise NotImplementedError

    def build_system_prompt_impl(
        self, *, ctx: RunContext[CtxT], call_id: str
    ) -> str | None:
        return self._prompt_builder.build_system_prompt_impl(ctx=ctx, call_id=call_id)

    def build_input_content_impl(
        self, in_args: InT, *, ctx: RunContext[CtxT], call_id: str
    ) -> Content:
        return self._prompt_builder.build_input_content_impl(
            in_args=in_args, ctx=ctx, call_id=call_id
        )

    def check_for_final_answer_impl(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        **kwargs: Any,
    ) -> str | None:
        return self._policy_executor.check_for_final_answer_impl(
            memory, ctx=ctx, call_id=call_id, **kwargs
        )

    async def on_before_generate_impl(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
        llm_settings: dict[str, Any],
    ) -> None:
        return await self._policy_executor.on_before_generate_impl(
            memory,
            ctx=ctx,
            call_id=call_id,
            num_turns=num_turns,
            llm_settings=llm_settings,
        )

    async def on_after_generate_impl(
        self,
        memory: LLMAgentMemory,
        *,
        ctx: RunContext[CtxT],
        call_id: str,
        num_turns: int,
        llm_settings: dict[str, Any],
    ) -> None:
        return await self._policy_executor.on_after_generate_impl(
            memory,
            ctx=ctx,
            call_id=call_id,
            num_turns=num_turns,
            llm_settings=llm_settings,
        )

    def tool_outputs_to_messages_impl(
        self,
        tool_outputs: Sequence[Any],
        tool_calls: Sequence[ToolCall],
        *,
        ctx: RunContext[CtxT],
        call_id: str,
    ):
        return self._policy_executor.tool_outputs_to_messages_impl(
            tool_outputs=tool_outputs,
            tool_calls=tool_calls,
            ctx=ctx,
            call_id=call_id,
        )

    # Decorators as an alternative to overriding methods

    def add_output_parser(
        self, func: OutputParser[InT, OutT, CtxT]
    ) -> OutputParser[InT, OutT, CtxT]:
        if self._used_default_llm_response_schema:
            self._policy_executor.response_schema = None
        self.parse_output_impl = func
        return func

    def add_memory_preparator(self, func: MemoryPreparator) -> MemoryPreparator:
        self.memory_preparator_impl = func
        return func

    def add_system_prompt_builder(
        self, func: SystemPromptBuilder[CtxT]
    ) -> SystemPromptBuilder[CtxT]:
        self._prompt_builder.build_system_prompt_impl = func
        return func

    def add_input_content_builder(
        self, func: InputContentBuilder[InT, CtxT]
    ) -> InputContentBuilder[InT, CtxT]:
        self._prompt_builder.build_input_content_impl = func
        return func

    def add_final_answer_checker(
        self, func: FinalAnswerChecker[CtxT]
    ) -> FinalAnswerChecker[CtxT]:
        self._policy_executor.check_for_final_answer_impl = func
        return func

    def add_before_generate_hook(
        self, func: BeforeGenerateHook[CtxT]
    ) -> BeforeGenerateHook[CtxT]:
        self._policy_executor.on_before_generate_impl = func
        return func

    def add_after_generate_hook(
        self, func: AfterGenerateHook[CtxT]
    ) -> AfterGenerateHook[CtxT]:
        self._policy_executor.on_after_generate_impl = func
        return func

    def add_tool_output_converter(
        self, func: ToolOutputConverter[CtxT]
    ) -> ToolOutputConverter[CtxT]:
        self._policy_executor.tool_outputs_to_messages_impl = func
        return func

    # When methods are overridden in subclasses, pass them to the components

    def _register_overridden_implementations(self) -> None:
        base_cls = LLMAgent[Any, Any, Any]

        # Prompt builder

        if is_method_overridden("build_system_prompt_impl", self, base_cls):
            self._prompt_builder.build_system_prompt_impl = (
                self.build_system_prompt_impl
            )

        if is_method_overridden("build_input_content_impl", self, base_cls):
            self._prompt_builder.build_input_content_impl = (
                self.build_input_content_impl
            )

        # Policy executor

        if is_method_overridden("check_for_final_answer_impl", self, base_cls):
            self._policy_executor.check_for_final_answer_impl = (
                self.check_for_final_answer_impl
            )

        if is_method_overridden("on_before_generate_impl", self, base_cls):
            self._policy_executor.on_before_generate_impl = self.on_before_generate_impl

        if is_method_overridden("on_after_generate_impl", self, base_cls):
            self._policy_executor.on_after_generate_impl = self.on_after_generate_impl

        if is_method_overridden("tool_outputs_to_messages_impl", self, base_cls):
            self._policy_executor.tool_outputs_to_messages_impl = (
                self.tool_outputs_to_messages_impl
            )



import time
from typing import Any
from uuid import uuid4

from litellm import ChatCompletionAnnotation as LiteLLMAnnotation
from litellm.types.utils import ChoiceLogprobs as LiteLLMChoiceLogprobs
from openai.types.chat.chat_completion import (
    ChoiceLogprobs as OpenAIChoiceLogprobs,
)
from openai.types.chat.chat_completion_chunk import (
    ChoiceLogprobs as OpenAIChunkChoiceLogprobs,
)
from openai.types.chat.chat_completion_token_logprob import (
    ChatCompletionTokenLogprob as OpenAITokenLogprob,
)
from pydantic import BaseModel, Field, ValidationError, field_validator

from ..errors import CombineCompletionChunksError
from .completion import Completion, FinishReason, Usage
from .message import (
    AssistantMessage,
    RedactedThinkingBlock,
    Role,
    ThinkingBlock,
    ToolCall,
)


class CompletionChunkDeltaToolCall(BaseModel):
    id: str | None
    index: int
    tool_name: str | None
    tool_arguments: str | None


class CompletionChunkDelta(BaseModel):
    content: str | None = None
    refusal: str | None = None
    role: Role | None = None
    tool_calls: list[CompletionChunkDeltaToolCall] | None = None
    reasoning_content: str | None = None
    thinking_blocks: list[ThinkingBlock | RedactedThinkingBlock] | None = None
    annotations: list[LiteLLMAnnotation] | None = None
    provider_specific_fields: dict[str, Any] | None = None

    @property
    def thinking_delta(self) -> "CompletionChunkDelta | None":
        return (
            CompletionChunkDelta(
                reasoning_content=self.reasoning_content,
                thinking_blocks=self.thinking_blocks,
                role=self.role,
                provider_specific_fields=self.provider_specific_fields,
            )
            if self.reasoning_content or self.thinking_blocks
            else None
        )

    @property
    def tool_call_deltas(self) -> "list[CompletionChunkDelta] | None":
        return (
            [
                CompletionChunkDelta(
                    tool_calls=[tool_call],
                    role=self.role,
                    provider_specific_fields=self.provider_specific_fields,
                )
                for tool_call in self.tool_calls
            ]
            if self.tool_calls
            else None
        )

    @property
    def response_delta(self) -> "CompletionChunkDelta | None":
        return (
            CompletionChunkDelta(
                content=self.content,
                role=self.role,
                provider_specific_fields=self.provider_specific_fields,
            )
            if self.content
            else None
        )

    @property
    def annotations_delta(self) -> "CompletionChunkDelta | None":
        return (
            CompletionChunkDelta(
                annotations=self.annotations,
                role=self.role,
                provider_specific_fields=self.provider_specific_fields,
            )
            if self.annotations
            else None
        )

    @property
    def refusal_delta(self) -> "CompletionChunkDelta | None":
        return (
            CompletionChunkDelta(
                refusal=self.refusal,
                role=self.role,
                provider_specific_fields=self.provider_specific_fields,
            )
            if self.refusal
            else None
        )


class CompletionChunk(BaseModel):
    id: str = Field(default_factory=lambda: str(uuid4())[:8])
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str | None
    name: str | None = None
    system_fingerprint: str | None = None
    usage: Usage | None = None

    # Removed choices to add delta directly to CompletionChunk
    delta: CompletionChunkDelta
    finish_reason: FinishReason | None
    logprobs: OpenAIChunkChoiceLogprobs | LiteLLMChoiceLogprobs | Any | None = None

    # LiteLLM-specific fields
    provider_specific_fields: dict[str, Any] | None = None
    response_ms: float | None = None
    hidden_params: dict[str, Any] | None = None

    def split_into_specialized(
        self,
    ) -> "list[CompletionChunk]":
        delta = self.delta

        specialized_chunks: list[CompletionChunk] = []

        thinking_delta = delta.thinking_delta
        tool_call_deltas = delta.tool_call_deltas
        response_delta = delta.response_delta
        annotations_delta = delta.annotations_delta
        refusal_delta = delta.refusal_delta

        if thinking_delta is not None:
            new_chunk = self.model_copy(update={"delta": thinking_delta})
            specialized_chunks.append(
                ThinkingChunk.model_validate(new_chunk.model_dump())
            )

        if tool_call_deltas:
            for delta_tool_call in tool_call_deltas:
                new_chunk = self.model_copy(update={"delta": delta_tool_call})
                specialized_chunks.append(
                    ToolCallChunk.model_validate(new_chunk.model_dump())
                )

        if response_delta is not None:
            new_chunk = self.model_copy(update={"delta": response_delta})
            specialized_chunks.append(
                ResponseChunk.model_validate(new_chunk.model_dump())
            )

        if annotations_delta is not None:
            new_chunk = self.model_copy(update={"delta": annotations_delta})
            specialized_chunks.append(
                AnnotationsChunk.model_validate(new_chunk.model_dump())
            )

        if refusal_delta is not None:
            new_chunk = self.model_copy(update={"delta": refusal_delta})
            specialized_chunks.append(
                RefusalChunk.model_validate(new_chunk.model_dump())
            )

        return specialized_chunks


class ResponseChunk(CompletionChunk):
    @field_validator("delta")
    @classmethod
    def validate_delta(cls, delta: CompletionChunkDelta) -> CompletionChunkDelta:
        if not delta.content:
            raise ValidationError("ResponseChunk must have content in deltas.")

        if (
            delta.reasoning_content is not None
            or delta.thinking_blocks is not None
            or delta.tool_calls is not None
            or delta.refusal is not None
            or delta.annotations is not None
        ):
            raise ValidationError(
                "ResponseChunk should not have reasoning content, thinking blocks, "
                "tool calls, refusal, or annotations in deltas."
            )

        return delta

    @property
    def response(self) -> str:
        assert self.delta.content
        return self.delta.content


class ThinkingChunk(CompletionChunk):
    @field_validator("delta")
    @classmethod
    def validate_delta(cls, delta: CompletionChunkDelta) -> CompletionChunkDelta:
        if not (delta.thinking_blocks or delta.reasoning_content):
            raise ValidationError(
                "ThinkingChunk must have reasoning content or "
                "at least one thinking block."
            )
        if (
            delta.content is not None
            or delta.tool_calls is not None
            or delta.refusal is not None
            or delta.annotations is not None
        ):
            raise ValidationError(
                "ThinkingChunk should not have content, tool calls, "
                "refusal, or annotations in deltas."
            )

        return delta

    @property
    def thinking(self) -> str | list[ThinkingBlock | RedactedThinkingBlock]:
        delta = self.delta
        if delta.reasoning_content:
            return delta.reasoning_content
        if delta.thinking_blocks:
            return delta.thinking_blocks
        raise ValueError("ThinkingChunk has no reasoning_content or thinking_blocks")


class ToolCallChunk(CompletionChunk):
    @field_validator("delta")
    @classmethod
    def validate_delta(cls, delta: CompletionChunkDelta) -> CompletionChunkDelta:
        if not delta.tool_calls:
            raise ValidationError("ToolCallChunk must have tool calls in deltas.")
        if len(delta.tool_calls) != 1:
            raise ValidationError(
                "ToolCallChunk must have exactly one tool call in deltas."
            )

        if (
            delta.reasoning_content is not None
            or delta.thinking_blocks is not None
            or delta.content is not None
            or delta.refusal is not None
            or delta.annotations is not None
        ):
            raise ValidationError(
                "ToolCallChunk should not have reasoning content, thinking blocks, "
                "content, refusal, or annotations in deltas."
            )

        return delta

    @property
    def tool_call(self) -> CompletionChunkDeltaToolCall:
        assert self.delta.tool_calls is not None
        return self.delta.tool_calls[0]


class AnnotationsChunk(CompletionChunk):
    @field_validator("delta")
    @classmethod
    def validate_annotations_chunk(
        cls, delta: CompletionChunkDelta
    ) -> CompletionChunkDelta:
        if not delta.annotations:
            raise ValidationError("AnnotationsChunk must have annotations in deltas.")

        if (
            delta.reasoning_content is not None
            or delta.thinking_blocks is not None
            or delta.content is not None
            or delta.tool_calls is not None
            or delta.refusal is not None
        ):
            raise ValidationError(
                "AnnotationsChunk should not have reasoning content, thinking blocks, "
                "content, tool calls, or refusal in deltas."
            )

        return delta

    @property
    def annotations(self) -> list[LiteLLMAnnotation]:
        assert self.delta.annotations is not None
        return self.delta.annotations


class RefusalChunk(CompletionChunk):
    @field_validator("delta")
    @classmethod
    def validate_refusal_chunk(
        cls, delta: CompletionChunkDelta
    ) -> CompletionChunkDelta:
        if not delta.refusal:
            raise ValidationError("RefusalChunk must have refusal in deltas.")

        if (
            delta.reasoning_content is not None
            or delta.thinking_blocks is not None
            or delta.content is not None
            or delta.tool_calls is not None
            or delta.annotations is not None
        ):
            raise ValidationError(
                "RefusalChunk should not have reasoning content, thinking blocks, "
                "content, tool calls, or annotations in deltas."
            )

        return delta

    @property
    def refusal(self) -> str | None:
        return self.delta.refusal


def combine_completion_chunks(chunks: list[CompletionChunk]) -> Completion:
    if not chunks:
        raise CombineCompletionChunksError(
            "Cannot combine an empty list of completion chunks."
        )

    model_list = {chunk.model for chunk in chunks}
    if len(model_list) > 1:
        raise CombineCompletionChunksError("All chunks must have the same model.")
    model = model_list.pop()

    name_list = {chunk.name for chunk in chunks}
    if len(name_list) > 1:
        raise CombineCompletionChunksError("All chunks must have the same name.")
    name = name_list.pop()

    system_fingerprints_list = {chunk.system_fingerprint for chunk in chunks}
    if len(system_fingerprints_list) > 1:
        raise CombineCompletionChunksError(
            "All chunks must have the same system fingerprint."
        )
    system_fingerprint = system_fingerprints_list.pop()

    created_list = [chunk.created for chunk in chunks]
    created = max(created_list)

    content: str = ""
    reasoning_content: str = ""
    refusal: str = ""

    delta_tool_calls: list[CompletionChunkDeltaToolCall] | None = None
    logp_contents: list[OpenAITokenLogprob] = []
    logp_refusals: list[OpenAITokenLogprob] = []
    thinking_blocks: list[ThinkingBlock | RedactedThinkingBlock] = []
    annotations: list[LiteLLMAnnotation] = []

    logprobs: OpenAIChoiceLogprobs | None = None
    finish_reason: FinishReason | None = None
    usage: Usage | None = None

    if chunks:
        last_chunk = chunks[-1]

        # Usage is found in the last completion chunk if requested
        usage = last_chunk.usage

        # Take the last finish reason
        finish_reason = last_chunk.finish_reason

        # NOTE: this won't work because tool calls need to be merged across chunks
        delta_tool_calls = last_chunk.delta.tool_calls

    for chunk in chunks:
        # Concatenate content and refusal tokens
        content += chunk.delta.content or ""
        reasoning_content += chunk.delta.reasoning_content or ""
        refusal += chunk.delta.refusal or ""

        # Concatenate logprobs for content and refusal tokens
        if chunk.logprobs is not None:
            logp_contents.extend(chunk.logprobs.content or [])  # type: ignore
            logp_refusals.extend(chunk.logprobs.refusal or [])  # type: ignore
            thinking_blocks.extend(chunk.delta.thinking_blocks or [])
            annotations.extend(chunk.delta.annotations or [])

    tool_calls: list[ToolCall] | None = []
    if delta_tool_calls is not None:
        for _tool_call in delta_tool_calls:
            if (
                _tool_call.id is None
                or _tool_call.tool_name is None
                or _tool_call.tool_arguments is None
            ):
                raise CombineCompletionChunksError(
                    "Completion chunk tool calls must have id, tool_name, "
                    "and tool_arguments set."
                )
            tool_calls.append(
                ToolCall(
                    id=_tool_call.id,
                    tool_name=_tool_call.tool_name,
                    tool_arguments=_tool_call.tool_arguments,
                )
            )

    message = AssistantMessage(
        name=name,
        content=content or "<empty>",
        reasoning_content=(reasoning_content or None),
        thinking_blocks=(thinking_blocks or None),
        annotations=(annotations or None),
        refusal=(refusal or None),
        tool_calls=(tool_calls or None),
    )

    if logp_contents or logp_refusals:
        logprobs = OpenAIChoiceLogprobs(content=logp_contents, refusal=logp_refusals)

    return Completion(
        model=model,
        name=name,
        created=created,
        system_fingerprint=system_fingerprint,
        message=message,
        finish_reason=finish_reason,
        logprobs=logprobs,
        usage=usage,
    )



import logging
from collections.abc import AsyncIterator, Sequence
from functools import partial
from typing import Any, Generic
from uuid import uuid4

from grasp_agents.tracing_decorators import workflow

from .errors import RunnerError
from .packet import Packet
from .packet_pool import END_PROC_NAME, START_PROC_NAME, PacketPool
from .processors.base_processor import BaseProcessor
from .run_context import CtxT, RunContext
from .typing.events import Event, ProcPacketOutputEvent, RunResultEvent
from .typing.io import OutT

logger = logging.getLogger(__name__)


class Runner(Generic[OutT, CtxT]):
    def __init__(
        self,
        entry_proc: BaseProcessor[Any, Any, Any, CtxT],
        procs: Sequence[BaseProcessor[Any, Any, Any, CtxT]],
        ctx: RunContext[CtxT] | None = None,
        name: str | None = None,
    ) -> None:
        if entry_proc not in procs:
            raise RunnerError(
                f"Entry processor {entry_proc.name} must be in the list of processors: "
                f"{', '.join(proc.name for proc in procs)}"
            )
        if sum(1 for proc in procs if END_PROC_NAME in (proc.recipients or [])) != 1:
            raise RunnerError(
                "There must be exactly one processor with recipient 'END'."
            )

        self._entry_proc = entry_proc
        self._procs = procs
        self._name = name or str(uuid4())[:6]
        self._ctx = ctx or RunContext[CtxT](state=None)  # type: ignore

    @property
    def name(self) -> str:
        return self._name

    @property
    def ctx(self) -> RunContext[CtxT]:
        return self._ctx

    def _generate_call_id(self, proc: BaseProcessor[Any, Any, Any, CtxT]) -> str | None:
        return self._name + "/" + proc._generate_call_id(call_id=None)  # type: ignore

    def _unpack_packet(
        self, packet: Packet[Any] | None
    ) -> tuple[Packet[Any] | None, Any | None]:
        if packet and packet.sender == START_PROC_NAME:
            return None, packet.payloads[0]
        return packet, None

    async def _packet_handler(
        self,
        packet: Packet[Any],
        *,
        proc: BaseProcessor[Any, Any, Any, CtxT],
        pool: PacketPool,
        **run_kwargs: Any,
    ) -> None:
        _in_packet, _chat_inputs = self._unpack_packet(packet)

        logger.info(f"\n[Running processor {proc.name}]\n")

        out_packet = await proc.run(
            chat_inputs=_chat_inputs,
            in_packet=_in_packet,
            ctx=self._ctx,
            call_id=self._generate_call_id(proc),
            **run_kwargs,
        )

        route = out_packet.uniform_routing or out_packet.routing
        logger.info(
            f"\n[Finished running processor {proc.name}]\n"
            f"Posting output packet to recipients: {route}\n"
        )

        await pool.post(out_packet)

    async def _packet_handler_stream(
        self,
        packet: Packet[Any],
        *,
        proc: BaseProcessor[Any, Any, Any, CtxT],
        pool: PacketPool,
        **run_kwargs: Any,
    ) -> None:
        _in_packet, _chat_inputs = self._unpack_packet(packet)

        logger.info(f"\n[Running processor {proc.name}]\n")

        out_packet: Packet[Any] | None = None
        async for event in proc.run_stream(
            chat_inputs=_chat_inputs,
            in_packet=_in_packet,
            ctx=self._ctx,
            call_id=self._generate_call_id(proc),
            **run_kwargs,
        ):
            if isinstance(event, ProcPacketOutputEvent):
                out_packet = event.data
            await pool.push_event(event)

        assert out_packet is not None

        route = out_packet.uniform_routing or out_packet.routing
        logger.info(
            f"\n[Finished running processor {proc.name}]\n"
            f"Posting output packet to recipients: {route}\n"
        )

        await pool.post(out_packet)

    @workflow(name="runner")  # type: ignore
    async def run(self, chat_inputs: Any = "start", **run_kwargs: Any) -> Packet[OutT]:
        async with PacketPool() as pool:
            for proc in self._procs:
                pool.register_packet_handler(
                    proc_name=proc.name,
                    handler=partial(
                        self._packet_handler,
                        proc=proc,
                        pool=pool,
                        **run_kwargs,
                    ),
                )
            start_packet = Packet[Any](
                sender=START_PROC_NAME,
                routing=[[self._entry_proc.name]],
                payloads=[chat_inputs],
            )
            await pool.post(start_packet)

            return await pool.final_result()

    @workflow(name="runner_run")  # type: ignore
    async def run_stream(
        self, chat_inputs: Any = "start", **run_kwargs: Any
    ) -> AsyncIterator[Event[Any]]:
        async with PacketPool() as pool:
            for proc in self._procs:
                pool.register_packet_handler(
                    proc_name=proc.name,
                    handler=partial(
                        self._packet_handler_stream,
                        proc=proc,
                        pool=pool,
                        **run_kwargs,
                    ),
                )

            start_packet = Packet[Any](
                sender=START_PROC_NAME,
                routing=[[self._entry_proc.name]],
                payloads=[chat_inputs],
            )
            await pool.post(start_packet)

            async for event in pool.stream_events():
                if isinstance(
                    event, ProcPacketOutputEvent
                ) and event.data.uniform_routing == [END_PROC_NAME]:
                    yield RunResultEvent(
                        data=event.data,
                        proc_name=event.proc_name,
                        call_id=event.call_id,
                    )
                else:
                    yield event
